"""
A semantically annotated document.
It doesn't have a specific structure as that depends on the output format (e.g. TEI).
"""
import dataclasses
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import (
    Callable,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast
)
from typing_extensions import Protocol

from sciencebeam_parser.document.layout_document import (
    EMPTY_BLOCK,
    LayoutBlock,
    LayoutGraphic,
    LayoutToken
)


class SemanticContentWrapper(ABC):
    def get_text(self) -> str:
        return ' '.join((
            block.text
            for block in self.iter_blocks()
        ))

    def __len__(self) -> int:
        return len(list(self.iter_blocks()))

    def get_short_semantic_content_repr(self):
        return '%s(%r)' % (type(self).__name__, self.get_text())

    @abstractmethod
    def iter_blocks(self) -> Iterable[LayoutBlock]:
        pass

    def iter_tokens(self) -> Iterable[LayoutToken]:
        return (
            token
            for block in self.iter_blocks()
            for token in block.iter_all_tokens()
        )

    @property
    def merged_block(self) -> LayoutBlock:
        return LayoutBlock.merge_blocks(self.iter_blocks())


@dataclass
class SemanticSimpleContentWrapper(SemanticContentWrapper):
    content: LayoutBlock = EMPTY_BLOCK

    layout_block: dataclasses.InitVar[LayoutBlock] = None

    def __post_init__(self, layout_block: Optional[LayoutBlock] = None):
        assert isinstance(self.content, LayoutBlock)
        if layout_block is not None:
            self.add_content(layout_block)

    def iter_blocks(self) -> Iterable[LayoutBlock]:
        return [self.content]

    def add_content(self, block: LayoutBlock):
        self.content = LayoutBlock(
            lines=self.content.lines + block.lines
        )


class SemanticTextContentWrapper(SemanticSimpleContentWrapper):
    pass


class SemanticContentFactoryProtocol(Protocol):
    def __call__(self, layout_block: LayoutBlock) -> SemanticContentWrapper:
        pass


EMPTY_CONTENT = SemanticSimpleContentWrapper()


T_SemanticContentWrapper = TypeVar('T_SemanticContentWrapper', bound=SemanticContentWrapper)


@dataclass
class SemanticMixedContentWrapper(SemanticContentWrapper):
    mixed_content: List[SemanticContentWrapper] = field(default_factory=list)
    content_id: Optional[str] = None
    layout_block: dataclasses.InitVar[LayoutBlock] = None

    def __post_init__(self, layout_block: Optional[LayoutBlock] = None):
        if layout_block is not None:
            self.add_block_content(layout_block)

    def __len__(self):
        return len(self.mixed_content)

    def __iter__(self) -> Iterator[SemanticContentWrapper]:
        return iter(self.mixed_content)

    def is_empty(self):
        return not self.mixed_content

    def iter_blocks(self) -> Iterable[LayoutBlock]:
        return (
            block
            for content in self.mixed_content
            for block in content.iter_blocks()
        )

    def add_block_content(self, block: LayoutBlock):
        self.add_content(SemanticTextContentWrapper(block))

    def add_content(self, content: SemanticContentWrapper):
        assert not isinstance(content, LayoutBlock)
        self.mixed_content.append(content)

    def add_content_and_return_content(
        self, content: T_SemanticContentWrapper
    ) -> T_SemanticContentWrapper:
        self.add_content(content)
        return content

    def iter_by_type(
        self, type_: Type[T_SemanticContentWrapper]
    ) -> Iterable[T_SemanticContentWrapper]:
        return (
            content for content in self.mixed_content
            if isinstance(content, type_)
        )

    def iter_by_type_recursively(
        self, type_: Type[T_SemanticContentWrapper]
    ) -> Iterable[T_SemanticContentWrapper]:
        return iter_by_semantic_type_recursively(self.mixed_content, type_)

    def iter_by_types_recursively(
        self, types_: Tuple[Type[T_SemanticContentWrapper], ...]
    ) -> Iterable[SemanticContentWrapper]:
        return iter_by_semantic_types_recursively(self.mixed_content, types_)

    def iter_parent_by_semantic_type_recursively(
        self, type_: Type[T_SemanticContentWrapper]
    ):
        return iter_parent_by_semantic_type_recursively(
            self.mixed_content, type_, self
        )

    def has_type(
        self, type_: Type[T_SemanticContentWrapper]
    ) -> bool:
        return next(iter(self.iter_by_type(type_)), None) is not None

    def view_by_type(self, type_: Type[T_SemanticContentWrapper]) -> 'SemanticMixedContentWrapper':
        return SemanticMixedContentWrapper(list(self.iter_by_type(type_)))

    def flat_map_inplace(
        self,
        fn: Callable[[SemanticContentWrapper], Sequence[SemanticContentWrapper]]
    ):
        self.mixed_content = [
            replaced_content
            for content in self.mixed_content
            for replaced_content in fn(content)
        ]

    def flat_map_inplace_by_type(
        self,
        type_: Type[T_SemanticContentWrapper],
        fn: Callable[[SemanticContentWrapper], Sequence[SemanticContentWrapper]]
    ):
        self.flat_map_inplace(
            lambda content: (
                fn(content) if isinstance(content, type_)
                else [content]
            )
        )

    def get_text_list(self) -> List[str]:
        return [content.get_text() for content in self.mixed_content]

    def get_text_by_type(self, type_: Type[T_SemanticContentWrapper]) -> str:
        return self.view_by_type(type_).get_text()


def iter_parent_by_semantic_type_recursively(
    semantic_content_iterable: Iterable[SemanticContentWrapper],
    type_: Type[T_SemanticContentWrapper],
    parent_content: SemanticContentWrapper
) -> Iterable[SemanticContentWrapper]:
    for semantic_content in semantic_content_iterable:
        if isinstance(semantic_content, type_):
            yield parent_content
            return
        if isinstance(semantic_content, SemanticMixedContentWrapper):
            yield from iter_parent_by_semantic_type_recursively(
                semantic_content.mixed_content,
                type_=type_,
                parent_content=semantic_content
            )


def iter_by_semantic_types_recursively(
    semantic_content_iterable: Iterable[SemanticContentWrapper],
    types_: Union[Type[T_SemanticContentWrapper], Tuple[Type[T_SemanticContentWrapper], ...]]
) -> Iterable[SemanticContentWrapper]:
    for semantic_content in semantic_content_iterable:
        if isinstance(semantic_content, types_):
            yield semantic_content
            continue
        if isinstance(semantic_content, SemanticMixedContentWrapper):
            yield from iter_by_semantic_types_recursively(
                semantic_content.mixed_content,
                types_=types_
            )


def iter_by_semantic_type_recursively(
    semantic_content_iterable: Iterable[SemanticContentWrapper],
    type_: Type[T_SemanticContentWrapper]
) -> Iterable[T_SemanticContentWrapper]:
    return cast(
        Iterable[T_SemanticContentWrapper],
        iter_by_semantic_types_recursively(
            semantic_content_iterable,
            type_
        )
    )


@dataclass
class SemanticNote(SemanticSimpleContentWrapper):
    note_type: str = 'other'


@dataclass
class SemanticMixedNote(SemanticMixedContentWrapper):
    note_type: str = 'other'


@dataclass
class SemanticOptionalValueSemanticMixedContentWrapper(SemanticMixedContentWrapper):
    value: Optional[str] = None


class SemanticHeading(SemanticMixedContentWrapper):
    pass


class SemanticParagraph(SemanticMixedContentWrapper):
    pass


class SemanticSectionTypes:
    BODY = 'BODY'
    BACK = 'BACK'
    ACKNOWLEDGEMENT = 'ACKNOWLEDGEMENT'
    OTHER = 'OTHER'


class SemanticLabel(SemanticSimpleContentWrapper):
    pass


class SemanticCaption(SemanticSimpleContentWrapper):
    pass


class SemanticTitle(SemanticSimpleContentWrapper):
    pass


class SemanticJournal(SemanticSimpleContentWrapper):
    pass


class SemanticVolume(SemanticSimpleContentWrapper):
    pass


class SemanticIssue(SemanticSimpleContentWrapper):
    pass


@dataclass
class SemanticPageRange(SemanticSimpleContentWrapper):
    from_page: Optional[str] = None
    to_page: Optional[str] = None


class SemanticPublisher(SemanticSimpleContentWrapper):
    pass


class SemanticLocation(SemanticSimpleContentWrapper):
    pass


@dataclass
class SemanticDate(SemanticSimpleContentWrapper):
    year: Optional[int] = None


class SemanticExternalIdentifierTypes:
    ARXIV = 'ARXIV'
    DOI = 'DOI'
    PII = 'PII'
    PMCID = 'PMCID'
    PMID = 'PMID'


@dataclass
class SemanticExternalIdentifier(SemanticSimpleContentWrapper):
    value: Optional[str] = None
    external_identifier_type: Optional[str] = None


class SemanticExternalUrl(SemanticOptionalValueSemanticMixedContentWrapper):
    pass


class SemanticAbstract(SemanticSimpleContentWrapper):
    pass


class SemanticRawNameList(SemanticMixedContentWrapper):
    pass


T_SemanticRawNameList = TypeVar('T_SemanticRawNameList', bound=SemanticRawNameList)


class SemanticRawAuthors(SemanticRawNameList):
    pass


class SemanticRawEditors(SemanticRawNameList):
    pass


class SemanticRawAffiliation(SemanticMixedContentWrapper):
    pass


class SemanticRawAddress(SemanticMixedContentWrapper):
    pass


class SemanticRawAffiliationAddress(SemanticMixedContentWrapper):
    pass


class SemanticMarker(SemanticSimpleContentWrapper):
    pass


class SemanticNamePart(SemanticOptionalValueSemanticMixedContentWrapper):
    pass


class SemanticNameTitle(SemanticNamePart):
    pass


class SemanticNameSuffix(SemanticNamePart):
    pass


class SemanticGivenName(SemanticNamePart):
    pass


class SemanticMiddleName(SemanticNamePart):
    pass


class SemanticSurname(SemanticNamePart):
    pass


class SemanticName(SemanticMixedContentWrapper):
    @property
    def label_text(self) -> str:
        return self.view_by_type(SemanticLabel).get_text()

    @property
    def given_name_text(self) -> str:
        return self.view_by_type(SemanticGivenName).get_text()

    @property
    def surname_text(self) -> str:
        return self.view_by_type(SemanticSurname).get_text()


T_SemanticName = TypeVar('T_SemanticName', bound=SemanticName)


class SemanticAuthor(SemanticName):
    pass


class SemanticEditor(SemanticName):
    pass


class SemanticInstitution(SemanticMixedContentWrapper):
    pass


class SemanticDepartment(SemanticMixedContentWrapper):
    pass


class SemanticLaboratory(SemanticMixedContentWrapper):
    pass


class SemanticAddressField(SemanticMixedContentWrapper):
    pass


class SemanticAddressLine(SemanticAddressField):
    pass


class SemanticPostCode(SemanticAddressField):
    pass


class SemanticPostBox(SemanticAddressField):
    pass


class SemanticRegion(SemanticAddressField):
    pass


class SemanticSettlement(SemanticAddressField):
    pass


class SemanticCountry(SemanticAddressField):
    pass


class SemanticAffiliationAddress(SemanticMixedContentWrapper):
    pass


class SemanticRawReferenceText(SemanticMixedContentWrapper):
    pass


class SemanticRawReference(SemanticMixedContentWrapper):
    pass


class SemanticReference(SemanticMixedContentWrapper):
    pass


class SemanticInvalidReference(SemanticMixedContentWrapper):
    pass


class SemanticReferenceList(SemanticMixedContentWrapper):
    pass


class SemanticRawFigure(SemanticMixedContentWrapper):
    pass


class SemanticFigure(SemanticMixedContentWrapper):
    pass


class SemanticRawTable(SemanticMixedContentWrapper):
    pass


class SemanticTable(SemanticMixedContentWrapper):
    pass


class SemanticRawEquationContent(SemanticMixedContentWrapper):
    pass


class SemanticRawEquation(SemanticMixedContentWrapper):
    pass


@dataclass
class SemanticGraphic(SemanticSimpleContentWrapper):
    layout_graphic: Optional[LayoutGraphic] = None
    relative_path: Optional[str] = None

    def get_short_semantic_content_repr(self):
        if not self.layout_graphic:
            return repr(self)
        return '%s(layout_graphic.local_file_path=%r)' % (
            type(self).__name__,
            self.layout_graphic.local_file_path
        )


@dataclass
class SemanticCitation(SemanticSimpleContentWrapper):
    target_content_id: Optional[str] = None


class SemanticFigureCitation(SemanticCitation):
    pass


class SemanticTableCitation(SemanticCitation):
    pass


class SemanticReferenceCitation(SemanticCitation):
    pass


class SemanticFront(SemanticMixedContentWrapper):
    @property
    def authors(self) -> List[SemanticAuthor]:
        return list(self.iter_by_type(SemanticAuthor))

    def get_raw_authors_text(self) -> str:
        return '\n'.join(self.view_by_type(SemanticRawAuthors).get_text_list())

    def get_authors_text(self) -> str:
        return '\n'.join(self.view_by_type(SemanticAuthor).get_text_list())


@dataclass
class SemanticSection(SemanticMixedContentWrapper):
    section_type: str = SemanticSectionTypes.OTHER

    @property
    def headings(self) -> List[SemanticHeading]:
        return list(self.iter_by_type(SemanticHeading))

    def get_heading_text(self) -> str:
        return '\n'.join(self.view_by_type(SemanticHeading).get_text_list())

    @property
    def paragraphs(self) -> List[SemanticParagraph]:
        return list(self.iter_by_type(SemanticParagraph))

    def get_paragraph_text_list(self) -> List[str]:
        return self.view_by_type(SemanticParagraph).get_text_list()

    def add_heading_block(self, block: LayoutBlock) -> SemanticHeading:
        return self.add_content_and_return_content(SemanticHeading(layout_block=block))

    def add_new_paragraph(self) -> SemanticParagraph:
        return self.add_content_and_return_content(SemanticParagraph())

    def add_note(self, block: LayoutBlock, note_type: str) -> SemanticNote:
        return self.add_content_and_return_content(
            SemanticNote(block, note_type=note_type)
        )

    def get_notes(self, note_type: str) -> List[SemanticNote]:
        return [
            note
            for note in self.iter_by_type(SemanticNote)
            if note.note_type == note_type
        ]

    def get_notes_text_list(self, note_type: str) -> List[str]:
        return [note.get_text() for note in self.get_notes(note_type)]

    @property
    def sections(self) -> List['SemanticSection']:
        return list(self.iter_by_type(SemanticSection))

    def get_sections(
        self,
        section_type: Optional[str] = None
    ) -> List['SemanticSection']:
        return [
            section
            for section in self.iter_by_type(SemanticSection)
            if not section_type or section.section_type == section_type
        ]

    def view_by_section_type(self, section_type: str) -> 'SemanticMixedContentWrapper':
        return SemanticMixedContentWrapper(
            cast(List[SemanticContentWrapper], self.get_sections(section_type))
        )

    def add_new_section(
        self,
        section_type: str = SemanticSectionTypes.OTHER
    ) -> 'SemanticSection':
        return self.add_content_and_return_content(
            SemanticSection(section_type=section_type)
        )


class SemanticDocument(SemanticMixedContentWrapper):
    def __init__(self):
        self.front = SemanticFront()
        self.body_section = SemanticSection(section_type=SemanticSectionTypes.BODY)
        self.back_section = SemanticSection(section_type=SemanticSectionTypes.BACK)
        super().__init__([self.front, self.body_section, self.back_section])



import logging

from sciencebeam_parser.models.training_data import (
    NO_NS_TEI_E,
    AbstractTeiTrainingDataGenerator,
    AbstractTrainingTeiParser
)


LOGGER = logging.getLogger(__name__)


# based on:
# https://github.com/kermitt2/grobid/blob/0.7.0/grobid-core/src/main/java/org/grobid/core/engines/ReferenceSegmenterParser.java

ROOT_TRAINING_XML_ELEMENT_PATH = ['text', 'listBibl']

TRAINING_XML_ELEMENT_PATH_BY_LABEL = {
    '<reference>': ROOT_TRAINING_XML_ELEMENT_PATH + ['bibl'],
    '<label>': ROOT_TRAINING_XML_ELEMENT_PATH + ['bibl', 'label']
}

RESET_TRAINING_XML_ELEMENT_PATH_BY_LABEL = {
    '<label>': ROOT_TRAINING_XML_ELEMENT_PATH
}


class ReferenceSegmenterTeiTrainingDataGenerator(AbstractTeiTrainingDataGenerator):
    DEFAULT_TEI_FILENAME_SUFFIX = '.references.referenceSegmenter.tei.xml'
    DEFAULT_DATA_FILENAME_SUFFIX = '.references.referenceSegmenter'

    def __init__(self):
        super().__init__(
            root_training_xml_element_path=ROOT_TRAINING_XML_ELEMENT_PATH,
            training_xml_element_path_by_label=TRAINING_XML_ELEMENT_PATH_BY_LABEL,
            reset_training_xml_element_path_by_label=RESET_TRAINING_XML_ELEMENT_PATH_BY_LABEL,
            element_maker=NO_NS_TEI_E,
            default_tei_filename_suffix=(
                ReferenceSegmenterTeiTrainingDataGenerator.DEFAULT_TEI_FILENAME_SUFFIX
            ),
            default_data_filename_suffix=(
                ReferenceSegmenterTeiTrainingDataGenerator.DEFAULT_DATA_FILENAME_SUFFIX
            ),
            default_tei_sub_directory='reference-segmenter/corpus/tei',
            default_data_sub_directory='reference-segmenter/corpus/raw'
        )


class ReferenceSegmenterTrainingTeiParser(AbstractTrainingTeiParser):
    def __init__(self) -> None:
        super().__init__(
            root_training_xml_element_path=ROOT_TRAINING_XML_ELEMENT_PATH,
            training_xml_element_path_by_label=(
                TRAINING_XML_ELEMENT_PATH_BY_LABEL
            ),
            use_tei_namespace=False
        )



import logging
import math
from abc import ABC, abstractmethod
from typing import Dict, Iterable, List, NamedTuple, Optional, Sequence, Set, cast

import PIL.Image

from sciencebeam_parser.utils.image import get_image_with_max_resolution
from sciencebeam_parser.document.layout_document import (
    LayoutBlock,
    LayoutGraphic,
    LayoutPageCoordinates,
    LayoutPageMeta
)
from sciencebeam_parser.document.semantic_document import (
    SemanticContentWrapper,
    SemanticGraphic,
    SemanticLabel,
    SemanticMixedContentWrapper
)
from sciencebeam_parser.ocr_models.ocr_model import OpticalCharacterRecognitionModel
from sciencebeam_parser.processors.ref_matching import SimpleContentIdMatcher


LOGGER = logging.getLogger(__name__)


DEFAULT_MAX_GRAPHIC_DISTANCE = 0.3


class GraphicMatch(NamedTuple):
    semantic_graphic: SemanticGraphic
    candidate_semantic_content: SemanticContentWrapper

    def __repr__(self) -> str:
        return '%s(semantic_graphic=%r, candidate_semantic_content=%s)' % (
            type(self).__name__,
            self.semantic_graphic.get_short_semantic_content_repr(),
            self.candidate_semantic_content.get_short_semantic_content_repr()
        )


class GraphicMatchResult(NamedTuple):
    graphic_matches: Sequence[GraphicMatch]
    unmatched_graphics: Sequence[SemanticGraphic]

    def __len__(self):
        return len(self.graphic_matches)

    def __iter__(self):
        return iter(self.graphic_matches)


class GraphicMatcher(ABC):
    @abstractmethod
    def get_graphic_matches(
        self,
        semantic_graphic_list: Sequence[SemanticGraphic],
        candidate_semantic_content_list: Sequence[SemanticContentWrapper]
    ) -> GraphicMatchResult:
        pass


class ChainedGraphicMatcher(GraphicMatcher):
    def __init__(self, graphic_matchers: Sequence[GraphicMatcher]):
        super().__init__()
        self.graphic_matchers = graphic_matchers

    def get_graphic_matches(
        self,
        semantic_graphic_list: Sequence[SemanticGraphic],
        candidate_semantic_content_list: Sequence[SemanticContentWrapper]
    ) -> GraphicMatchResult:
        current_graphic_match_result = GraphicMatchResult(
            graphic_matches=[],
            unmatched_graphics=semantic_graphic_list
        )
        for graphic_matcher in self.graphic_matchers:
            if not current_graphic_match_result.unmatched_graphics:
                break
            temp_graphic_matches = graphic_matcher.get_graphic_matches(
                current_graphic_match_result.unmatched_graphics,
                candidate_semantic_content_list
            )
            if not temp_graphic_matches.graphic_matches:
                continue
            current_graphic_match_result = GraphicMatchResult(
                graphic_matches=(
                    list(current_graphic_match_result.graphic_matches)
                    + list(temp_graphic_matches.graphic_matches)
                ),
                unmatched_graphics=temp_graphic_matches.unmatched_graphics
            )
        return current_graphic_match_result


class BoundingBoxDistance(NamedTuple):
    page_number_diff: int = 0
    delta_x: float = 0
    delta_y: float = 0
    euclidean_distance: float = 0

    def get_sort_key(self):
        return self.euclidean_distance

    def is_better_than(self, other: Optional['BoundingBoxDistance']) -> bool:
        if not other:
            return True
        return self.get_sort_key() < other.get_sort_key()


def get_bounding_box_distance(
    bounding_box_1: LayoutPageCoordinates,
    bounding_box_2: LayoutPageCoordinates
) -> BoundingBoxDistance:
    page_number_diff = abs(
        bounding_box_1.page_number - bounding_box_2.page_number
    )
    delta_x = max(
        0,
        bounding_box_2.x - (bounding_box_1.x + bounding_box_1.width),
        bounding_box_1.x - (bounding_box_2.x + bounding_box_2.width)
    )
    delta_y = max(
        0,
        bounding_box_2.y - (bounding_box_1.y + bounding_box_1.height),
        bounding_box_1.y - (bounding_box_2.y + bounding_box_2.height)
    )
    euclidean_distance = math.sqrt(delta_x**2 + delta_y**2)
    return BoundingBoxDistance(
        page_number_diff=page_number_diff,
        delta_x=delta_x,
        delta_y=delta_y,
        euclidean_distance=euclidean_distance
    )


def get_sorted_bounding_box_distances(
    bounding_box_distance_list: Iterable[BoundingBoxDistance]
) -> List[BoundingBoxDistance]:
    return sorted(
        bounding_box_distance_list,
        key=BoundingBoxDistance.get_sort_key
    )


def get_bounding_box_list_distance(
    bounding_box_list_1: Sequence[LayoutPageCoordinates],
    bounding_box_list_2: Sequence[LayoutPageCoordinates]
) -> BoundingBoxDistance:
    sorted_distances = get_sorted_bounding_box_distances([
        get_bounding_box_distance(bounding_box_1, bounding_box_2)
        for bounding_box_1 in bounding_box_list_1
        for bounding_box_2 in bounding_box_list_2
    ])
    return sorted_distances[0]


class BoundingBoxRef(NamedTuple):
    key: int
    bounding_box_list: Sequence[LayoutPageCoordinates]
    semantic_content: SemanticContentWrapper

    def __repr__(self) -> str:
        return '%s(key=%r, bounding_box_list=%r, semantic_content=%s)' % (
            type(self).__name__,
            self.key,
            self.bounding_box_list,
            self.semantic_content.get_short_semantic_content_repr()
        )

    def get_distance_to(self, other: 'BoundingBoxRef') -> BoundingBoxDistance:
        return get_bounding_box_list_distance(
            self.bounding_box_list,
            other.bounding_box_list
        )

    def with_extended_bounding_box_list(
        self,
        bounding_box_list: Sequence[LayoutPageCoordinates]
    ) -> 'BoundingBoxRef':
        return self._replace(  # pylint: disable=no-member
            bounding_box_list=list(self.bounding_box_list) + list(bounding_box_list)
        )


class BoundingBoxDistanceBetween(NamedTuple):
    bounding_box_distance: BoundingBoxDistance
    bounding_box_ref_1: BoundingBoxRef
    bounding_box_ref_2: BoundingBoxRef

    def get_sort_key(self):
        return self.bounding_box_distance.get_sort_key()

    def is_better_than(self, other: Optional['BoundingBoxDistanceBetween']) -> bool:
        if not other:
            return True
        return self.bounding_box_distance.is_better_than(other.bounding_box_distance)


def get_graphic_match_for_distance_between(
    distance_between: BoundingBoxDistanceBetween
) -> GraphicMatch:
    return GraphicMatch(
        semantic_graphic=cast(
            SemanticGraphic,
            distance_between.bounding_box_ref_1.semantic_content
        ),
        candidate_semantic_content=(
            distance_between.bounding_box_ref_2.semantic_content
        )
    )


def get_normalized_bounding_box_for_page_coordinates_and_page_meta(
    coordinates: LayoutPageCoordinates,
    page_meta: LayoutPageMeta
) -> LayoutPageCoordinates:
    page_coordinates = page_meta.coordinates
    assert page_coordinates is not None
    return LayoutPageCoordinates(
        x=coordinates.x / page_coordinates.width,
        y=(coordinates.y / page_coordinates.height) + page_meta.page_number,
        width=coordinates.width / page_coordinates.width,
        height=coordinates.height / page_coordinates.height,
        page_number=coordinates.page_number
    )


def get_normalized_bounding_box_list_for_layout_graphic(
    layout_graphic: LayoutGraphic
) -> Sequence[LayoutPageCoordinates]:
    if not layout_graphic.coordinates:
        return []
    return [get_normalized_bounding_box_for_page_coordinates_and_page_meta(
        layout_graphic.coordinates,
        page_meta=layout_graphic.page_meta
    )]


def get_normalized_bounding_box_list_for_layout_block(
    layout_block: LayoutBlock
) -> Sequence[LayoutPageCoordinates]:
    page_meta_by_page_number = {
        token.line_meta.page_meta.page_number: token.line_meta.page_meta
        for line in layout_block.lines
        for token in line.tokens
    }
    LOGGER.debug('page_meta_by_page_number: %r', page_meta_by_page_number)
    merged_coordinates_list = layout_block.get_merged_coordinates_list()
    return [
        get_normalized_bounding_box_for_page_coordinates_and_page_meta(
            coordinates=coordinates,
            page_meta=page_meta_by_page_number[coordinates.page_number]
        )
        for coordinates in merged_coordinates_list
    ]


def get_coordinates_sort_key(coordinates: LayoutPageCoordinates):
    return (coordinates.page_number, coordinates.y, coordinates.x)


def get_semantic_content_sort_key(semantic_content: SemanticContentWrapper):
    for token in semantic_content.iter_tokens():
        if token.coordinates:
            return get_coordinates_sort_key(token.coordinates)
    raise RuntimeError('no token with coordinates found')


def get_semantic_graphic_sort_key(semantic_graphic: SemanticGraphic):
    assert semantic_graphic.layout_graphic
    assert semantic_graphic.layout_graphic.coordinates
    return get_coordinates_sort_key(semantic_graphic.layout_graphic.coordinates)


def get_graphic_match_sort_key(graphic_match: GraphicMatch):
    return (
        get_semantic_content_sort_key(graphic_match.candidate_semantic_content),
        get_semantic_graphic_sort_key(graphic_match.semantic_graphic)
    )


def get_sorted_graphic_matches(graphic_matches: Sequence[GraphicMatch]) -> Sequence[GraphicMatch]:
    return sorted(graphic_matches, key=get_graphic_match_sort_key)


class _BoundingBoxDistanceGraphicMatcherInstance(NamedTuple):
    graphic_bounding_box_ref_list: Sequence[BoundingBoxRef]
    candidate_bounding_box_ref_list: Sequence[BoundingBoxRef]
    max_distance: float

    @staticmethod
    def create(
        semantic_graphic_list: Sequence[SemanticGraphic],
        candidate_semantic_content_list: Sequence[SemanticContentWrapper],
        ignored_graphic_types: Set[str],
        max_distance: float
    ) -> '_BoundingBoxDistanceGraphicMatcherInstance':
        graphic_bounding_box_ref_list = [
            BoundingBoxRef(
                id(semantic_graphic),
                bounding_box_list=get_normalized_bounding_box_list_for_layout_graphic(
                    semantic_graphic.layout_graphic
                ),
                semantic_content=semantic_graphic
            )
            for semantic_graphic in semantic_graphic_list
            if (
                semantic_graphic.layout_graphic
                and semantic_graphic.layout_graphic.coordinates
                and semantic_graphic.layout_graphic.graphic_type not in ignored_graphic_types
            )
        ]
        LOGGER.debug('graphic_bounding_box_ref_list: %r', graphic_bounding_box_ref_list)
        candidate_bounding_box_ref_list = [
            BoundingBoxRef(
                id(candidate_semantic_content),
                bounding_box_list=get_normalized_bounding_box_list_for_layout_block(
                    candidate_semantic_content
                    .merged_block
                ),
                semantic_content=candidate_semantic_content
            )
            for candidate_semantic_content in candidate_semantic_content_list
        ]
        LOGGER.debug('candidate_bounding_box_ref_list: %r', candidate_bounding_box_ref_list)
        return _BoundingBoxDistanceGraphicMatcherInstance(
            graphic_bounding_box_ref_list=graphic_bounding_box_ref_list,
            candidate_bounding_box_ref_list=candidate_bounding_box_ref_list,
            max_distance=max_distance
        )

    def is_accept_distance(self, distance_between: BoundingBoxDistanceBetween) -> bool:
        return distance_between.bounding_box_distance.euclidean_distance < self.max_distance

    def with_graphic_bounding_box_ref_list(
        self,
        graphic_bounding_box_ref_list: Sequence[BoundingBoxRef]
    ) -> '_BoundingBoxDistanceGraphicMatcherInstance':
        return self._replace(  # pylint: disable=no-member
            graphic_bounding_box_ref_list=graphic_bounding_box_ref_list
        )

    def with_candidate_bounding_box_ref_list(
        self,
        candidate_bounding_box_ref_list: Sequence[BoundingBoxRef]
    ) -> '_BoundingBoxDistanceGraphicMatcherInstance':
        return self._replace(  # pylint: disable=no-member
            candidate_bounding_box_ref_list=candidate_bounding_box_ref_list
        )

    def get_sorted_distances_between_for_graphic_bounding_box_ref(
        self,
        graphic_bounding_box_ref: BoundingBoxRef
    ) -> Sequence[BoundingBoxDistanceBetween]:
        return sorted(
            [
                BoundingBoxDistanceBetween(
                    bounding_box_distance=graphic_bounding_box_ref.get_distance_to(
                        candidate_bounding_box_ref
                    ),
                    bounding_box_ref_1=graphic_bounding_box_ref,
                    bounding_box_ref_2=candidate_bounding_box_ref
                )
                for candidate_bounding_box_ref in self.candidate_bounding_box_ref_list
            ],
            key=BoundingBoxDistanceBetween.get_sort_key
        )

    def get_best_distance_between_batch(self) -> Sequence[BoundingBoxDistanceBetween]:
        best_distance_between_by_candidate_key: Dict[int, BoundingBoxDistanceBetween] = {}
        for graphic_bounding_box_ref in self.graphic_bounding_box_ref_list:
            sorted_distances_between = (
                self.get_sorted_distances_between_for_graphic_bounding_box_ref(
                    graphic_bounding_box_ref
                )
            )
            if not sorted_distances_between:
                continue
            best_distance_between = sorted_distances_between[0]
            if not self.is_accept_distance(best_distance_between):
                LOGGER.debug('not accepting distance: %r', best_distance_between)
                continue
            LOGGER.debug('sorted_distances_between: %r', sorted_distances_between)
            candidate_key = best_distance_between.bounding_box_ref_2.key
            previous_best_distance_between = (
                best_distance_between_by_candidate_key.get(candidate_key)
            )
            if not best_distance_between.is_better_than(previous_best_distance_between):
                LOGGER.debug(
                    'found better previous best distance between: %r > %r',
                    previous_best_distance_between,
                    best_distance_between
                )
                continue
            if previous_best_distance_between:
                LOGGER.debug(
                    'found better best distance between: %r < %r',
                    previous_best_distance_between,
                    best_distance_between
                )
            LOGGER.debug('accept candidate: %r -> %r', candidate_key, best_distance_between)
            best_distance_between_by_candidate_key[
                candidate_key
            ] = best_distance_between
        return list(best_distance_between_by_candidate_key.values())

    def iter_remaining_candidate_bounding_box_refs(
        self,
        best_distance_between_batch: Iterable[BoundingBoxDistanceBetween]
    ) -> Iterable[BoundingBoxRef]:
        for best_distance_between in best_distance_between_batch:
            yield best_distance_between.bounding_box_ref_2.with_extended_bounding_box_list(
                best_distance_between.bounding_box_ref_1.bounding_box_list
            )

    def iter_graphic_matches(self) -> Iterable[GraphicMatch]:
        graphic_matches: List[GraphicMatch] = []
        matcher = self
        while matcher.graphic_bounding_box_ref_list:
            best_distance_between_batch = matcher.get_best_distance_between_batch()
            graphic_matches.extend([
                get_graphic_match_for_distance_between(distance_between)
                for distance_between in best_distance_between_batch
            ])
            matched_graphic_keys = {
                distance_between.bounding_box_ref_1.key
                for distance_between in best_distance_between_batch
            }
            LOGGER.debug('matched_graphic_keys: %r', matched_graphic_keys)
            remaining_graphic_bounding_box_ref_list = [
                graphic_bounding_box_ref
                for graphic_bounding_box_ref in matcher.graphic_bounding_box_ref_list
                if graphic_bounding_box_ref.key not in matched_graphic_keys
            ]
            LOGGER.debug(
                'remaining_graphic_bounding_box_ref_list: %r',
                remaining_graphic_bounding_box_ref_list
            )
            if (
                len(remaining_graphic_bounding_box_ref_list)
                == len(matcher.graphic_bounding_box_ref_list)
            ):
                # not matched anything in this round
                break
            remaining_candidate_bounding_box_ref_list = list(
                matcher.iter_remaining_candidate_bounding_box_refs(
                    best_distance_between_batch
                )
            )
            LOGGER.debug(
                'remaining_candidate_bounding_box_ref_list: %r',
                remaining_candidate_bounding_box_ref_list
            )
            matcher = matcher.with_candidate_bounding_box_ref_list(
                remaining_candidate_bounding_box_ref_list
            ).with_graphic_bounding_box_ref_list(remaining_graphic_bounding_box_ref_list)
        return get_sorted_graphic_matches(graphic_matches)


class BoundingBoxDistanceGraphicMatcher(GraphicMatcher):
    def __init__(self, max_distance: float = DEFAULT_MAX_GRAPHIC_DISTANCE):
        super().__init__()
        # we ignore svgs for now because they currently represent the whole page
        # rather than individual images
        self.ignored_graphic_types = {'svg'}
        self.max_distance = max_distance

    def get_graphic_matches(
        self,
        semantic_graphic_list: Sequence[SemanticGraphic],
        candidate_semantic_content_list: Sequence[SemanticContentWrapper]
    ) -> GraphicMatchResult:
        matcher = _BoundingBoxDistanceGraphicMatcherInstance.create(
            semantic_graphic_list=semantic_graphic_list,
            candidate_semantic_content_list=candidate_semantic_content_list,
            ignored_graphic_types=self.ignored_graphic_types,
            max_distance=self.max_distance
        )
        graphic_matches = list(matcher.iter_graphic_matches())
        all_matched_graphic_keys = {
            id(graphic_match.semantic_graphic)
            for graphic_match in graphic_matches
        }
        unmatched_graphics = [
            semantic_graphic
            for semantic_graphic in semantic_graphic_list
            if id(semantic_graphic) not in all_matched_graphic_keys
        ]
        return GraphicMatchResult(
            graphic_matches=graphic_matches,
            unmatched_graphics=unmatched_graphics
        )


class AbstractGraphicTextGraphicMatcher(GraphicMatcher):
    @abstractmethod
    def get_text_for_semantic_graphic(self, semantic_graphic: SemanticGraphic) -> str:
        pass

    def get_text_for_candidate_semantic_content(
        self,
        semantic_content: SemanticContentWrapper
    ) -> str:
        assert isinstance(semantic_content, SemanticMixedContentWrapper)
        return semantic_content.get_text_by_type(SemanticLabel)

    def get_graphic_matches(
        self,
        semantic_graphic_list: Sequence[SemanticGraphic],
        candidate_semantic_content_list: Sequence[SemanticContentWrapper]
    ) -> GraphicMatchResult:
        graphic_text_list = [
            self.get_text_for_semantic_graphic(semantic_graphic)
            for semantic_graphic in semantic_graphic_list
        ]
        LOGGER.debug('graphic_text_list: %r', graphic_text_list)
        candidate_label_text_list = [
            self.get_text_for_candidate_semantic_content(candidate_semantic_content)
            for candidate_semantic_content in candidate_semantic_content_list
        ]
        LOGGER.debug('candidate_label_text_list: %r', candidate_label_text_list)
        content_id_matcher = SimpleContentIdMatcher(
            {str(index): value for index, value in enumerate(candidate_label_text_list)},
            prefix_length=3
        )
        graphic_matches: List[GraphicMatch] = []
        unmatched_graphics: List[SemanticGraphic] = []
        for graphic_index, graphic_text in enumerate(graphic_text_list):
            semantic_graphic = semantic_graphic_list[graphic_index]
            candidate_id: Optional[str] = None
            for graphic_text_line in graphic_text.splitlines():
                if not graphic_text_line:
                    continue
                candidate_id = content_id_matcher.get_id_by_text(graphic_text_line)
                if candidate_id is not None:
                    break
            LOGGER.debug('get_id_by_text: candidate_id=%r (text=%r)', candidate_id, graphic_text)
            if candidate_id is None:
                unmatched_graphics.append(semantic_graphic)
                continue
            candidate_semantic_content = candidate_semantic_content_list[int(candidate_id)]
            graphic_matches.append(GraphicMatch(
                semantic_graphic=semantic_graphic,
                candidate_semantic_content=candidate_semantic_content
            ))
        return GraphicMatchResult(
            graphic_matches=graphic_matches,
            unmatched_graphics=unmatched_graphics
        )


class GraphicRelatedBlockTextGraphicMatcher(AbstractGraphicTextGraphicMatcher):
    def get_text_for_semantic_graphic(self, semantic_graphic: SemanticGraphic) -> str:
        assert semantic_graphic.layout_graphic
        if semantic_graphic.layout_graphic.related_block:
            return '\n'.join([
                line.text.strip()
                for line in semantic_graphic.layout_graphic.related_block.lines
            ])
        return ''


DEFAULT_OCR_MAX_RESOLUTION = 1024


class OpticalCharacterRecognitionGraphicMatcher(AbstractGraphicTextGraphicMatcher):
    def __init__(
        self,
        ocr_model: OpticalCharacterRecognitionModel,
        max_resolution: int = DEFAULT_OCR_MAX_RESOLUTION
    ):
        super().__init__()
        self.ocr_model = ocr_model
        self.max_resolution = max_resolution

    def get_text_for_semantic_graphic(self, semantic_graphic: SemanticGraphic) -> str:
        assert semantic_graphic.layout_graphic
        if semantic_graphic.layout_graphic.graphic_type == 'svg':
            return ''
        if not semantic_graphic.layout_graphic.local_file_path:
            # images generated by CV object detection may not have a local file path
            # if we don't output assets
            LOGGER.info(
                'no local image found for layout graphic: %r',
                semantic_graphic.layout_graphic
            )
            return ''
        with PIL.Image.open(semantic_graphic.layout_graphic.local_file_path) as image:
            return self.ocr_model.predict_single(
                get_image_with_max_resolution(image, self.max_resolution)
            ).get_text()



import logging

from lxml import etree

from sciencebeam_parser.document.semantic_document import (
    SemanticContentWrapper,
    SemanticMixedNote
)
from sciencebeam_parser.document.tei.common import (
    TEI_E
)
from sciencebeam_parser.document.tei.factory import (
    SingleElementTeiElementFactory,
    TeiElementFactoryContext
)


LOGGER = logging.getLogger(__name__)


class SemanticMixedNoteTeiElementFactory(SingleElementTeiElementFactory):
    def get_tei_element_for_semantic_content(
        self,
        semantic_content: SemanticContentWrapper,
        context: TeiElementFactoryContext
    ) -> etree.ElementBase:
        LOGGER.debug('semantic_content: %s', semantic_content)
        assert isinstance(semantic_content, SemanticMixedNote)
        semantic_mixed_note = semantic_content
        note_type = semantic_mixed_note.note_type or 'other'
        children = [context.get_default_attributes_for_semantic_content(semantic_mixed_note)]
        children.append({'type': note_type})
        for child_semantic_content in semantic_mixed_note:
            children.extend(context.get_tei_child_elements_for_semantic_content(
                child_semantic_content
            ))
        return TEI_E('note', *children)



"""
Constants and functionality related to Media Types (formerly known as MIME types)

See:
  https://www.iana.org/assignments/media-types/media-types.xhtml
"""
import mimetypes
from typing import Optional, Sequence


class MediaTypes:
    """
    Media Types used by ScienceBeam Parser.
    Where possible, these correspond to official media types.
    In some instances, no official media type is defined yet.
    """
    PDF = 'application/pdf'
    DOC = 'application/msword'
    DOCX = 'application/vnd.openxmlformats-officedocument.wordprocessingml.document'
    DOTX = 'application/vnd.openxmlformats-officedocument.wordprocessingml.template'
    RTF = 'application/rtf'
    XML = 'application/xml'
    ZIP = 'application/zip'
    TEI_XML = 'application/tei+xml'
    JATS_XML = 'application/vnd.jats+xml'
    TEI_ZIP = 'application/tei+xml+zip'
    JATS_ZIP = 'application/vnd.jats+xml+zip'
    ALTO_XML = 'application/vnd.alto+xml'
    JSON = 'application/json'
    OCTET_STREAM = 'application/octet-stream'


WILDCARD_MEDIA_TYPE = '*/*'


MEDIA_TYPE_SUFFIX_MAP = {
    # fixed mime type suffix map (which may be incorrectly defined in Python 3.5)
    MediaTypes.DOC: '.doc',
    # additional types
    MediaTypes.TEI_XML: '.tei.xml',
    MediaTypes.JATS_XML: '.jats.xml',
    MediaTypes.TEI_ZIP: '.tei.zip',
    MediaTypes.JATS_ZIP: '.jats.zip'
}


def guess_extension_for_media_type(media_type: str) -> Optional[str]:
    ext = MEDIA_TYPE_SUFFIX_MAP.get(media_type)
    if not ext:
        ext = mimetypes.guess_extension(media_type)
    return ext


def guess_media_type_for_filename(filename: str) -> Optional[str]:
    return mimetypes.guess_type(filename)[0]


def get_first_matching_media_type(
    accept_media_types: Sequence[str],
    available_media_types: Sequence[str]
) -> Optional[str]:
    if not available_media_types:
        return None
    if not accept_media_types:
        return available_media_types[0]
    for accept_media_type in accept_media_types:
        if accept_media_type == WILDCARD_MEDIA_TYPE:
            return available_media_types[0]
        for available_media_type in available_media_types:
            if accept_media_type == available_media_type:
                return available_media_type
    return None



import os
import logging
from time import monotonic
from typing import Dict, Iterable, Mapping, Optional, Sequence, Set

import PIL.Image

from sciencebeam_parser.utils.bounding_box import BoundingBox
from sciencebeam_parser.document.semantic_document import SemanticGraphic
from sciencebeam_parser.document.layout_document import (
    DEFAULT_LAYOUT_PAGE_META,
    LayoutDocument,
    LayoutGraphic,
    LayoutPage,
    LayoutPageCoordinates
)
from sciencebeam_parser.cv_models.cv_model import ComputerVisionModel
from sciencebeam_parser.processors.document_page_image import DocumentPageImage
from sciencebeam_parser.processors.graphic_provider import (
    DocumentGraphicProvider,
    get_semantic_graphic_for_layout_graphic,
    get_semantic_graphic_list_for_layout_graphic_list
)


LOGGER = logging.getLogger(__name__)


def get_cropped_image(image: PIL.Image.Image, bounding_box: BoundingBox) -> PIL.Image.Image:
    return image.crop((
        bounding_box.x,
        bounding_box.y,
        bounding_box.right,
        bounding_box.bottom
    ))


def get_bounding_box_intersection_area_ratio(
    bounding_box_1: BoundingBox,
    bounding_box_2: BoundingBox,
    empty_ratio: float = 0.0
) -> float:
    max_area = max(bounding_box_1.area, bounding_box_2.area)
    if not max_area:
        return empty_ratio
    intersection_area = bounding_box_1.intersection(bounding_box_2).area
    return intersection_area / max_area


def get_layout_graphic_with_similar_coordinates(
    page_graphics: Sequence[LayoutGraphic],
    bounding_box: BoundingBox,
    threshold: float = 0.80,
    ignored_graphic_types: Set[str] = None
) -> Optional[LayoutGraphic]:
    sorted_area_intersection_bounding_boxes = sorted((
        (
            get_bounding_box_intersection_area_ratio(
                bounding_box,
                graphic.coordinates.bounding_box
            ),
            graphic
        )
        for graphic in page_graphics
        if graphic.coordinates and (
            not ignored_graphic_types or graphic.graphic_type not in ignored_graphic_types
        )
    ), key=lambda t: -t[0])
    if not sorted_area_intersection_bounding_boxes:
        return None
    LOGGER.debug(
        'sorted_area_intersection_bounding_boxes: %r',
        sorted_area_intersection_bounding_boxes
    )
    best_area_ratio, best_matching_graphic = sorted_area_intersection_bounding_boxes[0]
    if best_area_ratio < threshold:
        LOGGER.debug('best_area_ratio below threshold: %.3f < %.3f', best_area_ratio, threshold)
        return None
    return best_matching_graphic


class ComputerVisionDocumentGraphicProvider(DocumentGraphicProvider):
    def __init__(
        self,
        computer_vision_model: ComputerVisionModel,
        page_image_iterable: Iterable[DocumentPageImage],
        temp_dir: str
    ):
        super().__init__()
        self.computer_vision_model = computer_vision_model
        self.page_image_iterable = page_image_iterable
        self.temp_dir = temp_dir
        # ignoring svg for now because we are also ignoring it when matching graphics
        # an svg image may also not be standalone and require text to be complete
        self.ignored_graphic_types = {'svg'}

    def iter_semantic_graphic_for_image(  # pylint: disable=too-many-locals
        self,
        image: PIL.Image.Image,
        extract_graphic_assets: bool,
        page_number: int,
        page: Optional[LayoutPage]
    ) -> Iterable[SemanticGraphic]:
        LOGGER.debug('image size: %d x %d', image.width, image.height)
        page_meta = page.meta if page is not None else DEFAULT_LAYOUT_PAGE_META
        page_coordinates = (
            page.meta.coordinates if page is not None else None
        )
        page_graphics = (
            page.graphics if page is not None else []
        )
        cv_start = monotonic()
        cv_result = self.computer_vision_model.predict_single(image)
        cv_end = monotonic()
        cv_instances = cv_result.get_instances_by_type_names(['Figure', 'Table'])
        cv_type_name_and_coordinates_list = [
            (instance.get_type_name(), instance.get_bounding_box())
            for instance in cv_instances
        ]
        LOGGER.info(
            (
                'cv result, took=%.3fs, page_number=%d, image_size=%dx%d'
                ', cv_type_name_and_coordinates_list=%r'
            ),
            cv_end - cv_start,
            page_number,
            image.width,
            image.height,
            cv_type_name_and_coordinates_list
        )
        count_by_type_name_map: Dict[str, int] = {}
        for type_name, cv_coordinates in cv_type_name_and_coordinates_list:
            lower_type_name = type_name.lower()
            count_by_type_name_map[type_name] = count_by_type_name_map.get(type_name, 0) + 1
            item_number = count_by_type_name_map[type_name]
            local_image_path: Optional[str] = None
            relative_image_path: Optional[str] = None
            scaled_item_coordinates = cv_coordinates
            if page_coordinates:
                scaled_item_coordinates = (
                    cv_coordinates
                    .scale_by(
                        page_coordinates.width / image.width,
                        page_coordinates.height / image.height
                    )
                )
            matching_layout_graphic = get_layout_graphic_with_similar_coordinates(
                page_graphics=page_graphics,
                bounding_box=scaled_item_coordinates,
                ignored_graphic_types=self.ignored_graphic_types
            )
            if matching_layout_graphic is not None:
                yield get_semantic_graphic_for_layout_graphic(
                    matching_layout_graphic,
                    extract_graphic_assets=extract_graphic_assets
                )
                continue
            if extract_graphic_assets:
                local_image_path = os.path.join(
                    self.temp_dir, f'{lower_type_name}-{page_number}-{item_number}.png'
                )
                relative_image_path = os.path.basename(local_image_path)
                cropped_image = get_cropped_image(image, cv_coordinates)
                cropped_image.save(local_image_path)
            layout_graphic = LayoutGraphic(
                coordinates=LayoutPageCoordinates(
                    x=scaled_item_coordinates.x,
                    y=scaled_item_coordinates.y,
                    width=scaled_item_coordinates.width,
                    height=scaled_item_coordinates.height,
                    page_number=page_number
                ),
                page_meta=page_meta,
                graphic_type=f'cv-{lower_type_name}',
                local_file_path=local_image_path
            )
            semantic_graphic = SemanticGraphic(
                layout_graphic=layout_graphic,
                relative_path=relative_image_path
            )
            yield semantic_graphic

    def get_page_by_page_number_map(
        self,
        layout_document: LayoutDocument
    ) -> Mapping[int, Optional[LayoutPage]]:
        return {
            page.meta.page_number: page
            for page in layout_document.pages
        }

    def iter_semantic_graphic_for_layout_document(
        self,
        layout_document: LayoutDocument,
        extract_graphic_assets: bool
    ) -> Iterable[SemanticGraphic]:
        page_by_page_number_map = self.get_page_by_page_number_map(
            layout_document
        )
        LOGGER.debug(
            'cv model: page_by_page_number_map=%r',
            page_by_page_number_map
        )
        has_cv_semantic_graphic: bool = False
        for page_image in self.page_image_iterable:
            LOGGER.debug('page_image: %r', page_image)
            page_number = page_image.page_number
            with PIL.Image.open(page_image.page_image_path) as image:
                for semantic_graphic in self.iter_semantic_graphic_for_image(
                    image,
                    extract_graphic_assets=extract_graphic_assets,
                    page_number=page_number,
                    page=page_by_page_number_map.get(page_number)
                ):
                    has_cv_semantic_graphic = True
                    yield semantic_graphic
        if not has_cv_semantic_graphic:
            LOGGER.info('no graphics detected using cv model, falling back to regular graphics')
            yield from get_semantic_graphic_list_for_layout_graphic_list(
                layout_document.iter_all_graphics(),
                extract_graphic_assets=extract_graphic_assets
            )



import logging
import re
from abc import ABC, abstractmethod
from collections import Counter, defaultdict
from typing import Dict, List, Mapping, Optional, Sequence

from sciencebeam_parser.utils.tokenizer import iter_tokenized_tokens


LOGGER = logging.getLogger(__name__)


class ContentIdMatcher(ABC):
    @abstractmethod
    def get_id_by_text(self, text: str) -> Optional[str]:
        pass


def get_normalized_key_text(text: str):
    return re.sub(
        r'[^a-z0-9]',
        '',
        text.lower()
    )


def get_token_prefix_normalized_key_text(text: str, prefix_length: int = 1):
    return ''.join([
        get_normalized_key_text(
            token if re.search(r'\d', token) else token[:prefix_length]
        )
        for token in re.split(r'\s', text)
    ])


def get_normalized_key_tokens(text: str):
    return [
        get_normalized_key_text(token)
        for token in iter_tokenized_tokens(text)
        if token.strip()
    ]


class SimpleContentIdMatcher(ContentIdMatcher):
    def __init__(self, text_by_content_id: Mapping[str, str], prefix_length: int = 1):
        self.text_by_content_id = text_by_content_id
        self.content_id_by_text = {
            get_normalized_key_text(value): key
            for key, value in text_by_content_id.items()
        }
        self.content_id_by_token_prefix_text = {
            get_token_prefix_normalized_key_text(value, prefix_length=prefix_length): key
            for key, value in text_by_content_id.items()
        }
        self.prefix_length = prefix_length

    def get_id_by_text(self, text: str) -> Optional[str]:
        content_id = self.content_id_by_text.get(get_normalized_key_text(text))
        if content_id:
            return content_id
        content_id = self.content_id_by_token_prefix_text.get(
            get_token_prefix_normalized_key_text(text, prefix_length=self.prefix_length)
        )
        return content_id


class PartialContentIdMatcher(ContentIdMatcher):
    def __init__(self, text_by_content_id: Mapping[str, str]):
        self.content_ids_by_token_text: Dict[str, List[str]] = defaultdict(list)
        for content_id, text in text_by_content_id.items():
            for token in get_normalized_key_tokens(text):
                self.content_ids_by_token_text[token].append(content_id)

    def get_id_by_text(self, text: str) -> Optional[str]:
        tokens = get_normalized_key_tokens(text)
        LOGGER.debug('tokens: %r (text: %r)', tokens, text)
        if not tokens:
            return None
        content_id_counts = Counter((
            content_id
            for token in tokens
            for content_id in self.content_ids_by_token_text.get(token, [])
        ))
        LOGGER.debug('content_id_counts: %s', content_id_counts)
        if not content_id_counts:
            return None
        keys = list(content_id_counts.keys())
        if (
            len(content_id_counts) >= 2
            and content_id_counts[keys[0]] == content_id_counts[keys[1]]
        ):
            return None
        return keys[0]


class ChainedContentIdMatcher(ContentIdMatcher):
    def __init__(self, content_id_matchers: Sequence[ContentIdMatcher]):
        self.content_id_matchers = content_id_matchers

    def get_id_by_text(self, text: str) -> Optional[str]:
        for content_id_matcher in self.content_id_matchers:
            content_id = content_id_matcher.get_id_by_text(text)
            if content_id:
                return content_id
        return None



import logging
from typing import Iterable, Tuple

from sciencebeam_parser.document.layout_document import LayoutBlock
from sciencebeam_parser.document.semantic_document import SemanticDocument
from sciencebeam_parser.models.data import (
    DocumentFeaturesContext
)
from sciencebeam_parser.models.header.data import HeaderDataGenerator
from sciencebeam_parser.models.header.extract import HeaderSemanticExtractor
from sciencebeam_parser.models.header.training_data import (
    HeaderTeiTrainingDataGenerator,
    HeaderTrainingTeiParser
)
from sciencebeam_parser.models.model import Model


LOGGER = logging.getLogger(__name__)


class HeaderModel(Model):
    def get_data_generator(
        self,
        document_features_context: DocumentFeaturesContext
    ) -> HeaderDataGenerator:
        return HeaderDataGenerator(
            document_features_context=document_features_context
        )

    def get_semantic_extractor(self) -> HeaderSemanticExtractor:
        return HeaderSemanticExtractor()

    def get_tei_training_data_generator(self) -> HeaderTeiTrainingDataGenerator:
        return HeaderTeiTrainingDataGenerator()

    def get_training_tei_parser(self) -> HeaderTrainingTeiParser:
        return HeaderTrainingTeiParser()

    def update_semantic_document_with_entity_blocks(
        self,
        document: SemanticDocument,
        entity_tokens: Iterable[Tuple[str, LayoutBlock]]
    ):
        semantic_content_iterable = (
            self.get_semantic_extractor()
            .iter_semantic_content_for_entity_blocks(entity_tokens)
        )
        front = document.front
        for semantic_content in semantic_content_iterable:
            front.add_content(semantic_content)



import os
import pkg_resources


def get_office_script_directory():
    try:
        # with Apache Beam, __file__ might refer to the original file
        # rather than where it is installed to
        dist = pkg_resources.get_distribution("sciencebeam")
        return os.path.join(dist.location, 'sciencebeam/transformers/office_scripts')
    except pkg_resources.DistributionNotFound:
        # fallback to use __file__
        return os.path.dirname(__file__)



from abc import ABC, abstractmethod
from dataclasses import dataclass
import logging
from typing import Iterable, List, Mapping, NamedTuple, Optional, Sequence, Tuple, TypeVar, Union

from lxml import etree
from lxml.builder import ElementMaker

from sciencebeam_parser.utils.xml_writer import XmlTreeWriter
from sciencebeam_parser.utils.labels import get_split_prefix_label
from sciencebeam_parser.utils.tokenizer import get_tokenized_tokens
from sciencebeam_parser.document.tei.common import TEI_E, TEI_NS_PREFIX, tei_xpath
from sciencebeam_parser.document.layout_document import (
    LayoutLine,
    LayoutLineMeta,
    LayoutToken
)
from sciencebeam_parser.models.data import (
    NEW_DOCUMENT_MARKER,
    LabeledLayoutModelData,
    LabeledLayoutToken,
    LayoutModelData,
    NewDocumentMarker
)


LOGGER = logging.getLogger(__name__)


NO_NS_TEI_E = ElementMaker()


OTHER_LABELS = {'<other>', 'O'}


class ExtractInstruction:
    pass


class NewLineExtractInstruction(ExtractInstruction):
    pass


@dataclass
class ResetExtractInstruction(ExtractInstruction):
    reset_element_path: List[str]


def get_model_data_label(model_data: LayoutModelData) -> Optional[str]:
    if isinstance(model_data, LabeledLayoutModelData):
        return model_data.label
    return None


def is_same_layout_line(
    layout_line_1: Optional[LayoutLine],
    layout_line_2: Optional[LayoutLine]
) -> bool:
    assert layout_line_1 is not None
    assert layout_line_2 is not None
    return id(layout_line_1) == id(layout_line_2)


def is_same_model_data_layout_line(
    model_data_1: LayoutModelData,
    model_data_2: LayoutModelData
) -> bool:
    return is_same_layout_line(model_data_1.layout_line, model_data_2.layout_line)


def iter_group_model_data_by_line(
    model_data_iterable: Iterable[LayoutModelData]
) -> Iterable[Sequence[LayoutModelData]]:
    line_model_data_list: List[LayoutModelData] = []
    for model_data in model_data_iterable:
        if not line_model_data_list:
            line_model_data_list.append(model_data)
            continue
        previous_model_data = line_model_data_list[-1]
        if is_same_model_data_layout_line(
            model_data,
            previous_model_data
        ):
            LOGGER.debug('same line: %r - %r', model_data, previous_model_data)
            line_model_data_list.append(model_data)
            continue
        yield line_model_data_list
        line_model_data_list = [model_data]
    if line_model_data_list:
        yield line_model_data_list


def iter_model_data_with_new_line_instruction(
    model_data_iterable: Iterable[LayoutModelData]
) -> Iterable[Union[LayoutModelData, ExtractInstruction]]:
    line_model_data_list: List[LayoutModelData] = []
    for model_data in model_data_iterable:
        if not line_model_data_list:
            line_model_data_list.append(model_data)
            continue
        previous_model_data = line_model_data_list[-1]
        if is_same_model_data_layout_line(
            model_data,
            previous_model_data
        ):
            LOGGER.debug('same line: %r - %r', model_data, previous_model_data)
            line_model_data_list.append(model_data)
            continue
        yield from line_model_data_list
        yield NewLineExtractInstruction()
        line_model_data_list = [model_data]
    if line_model_data_list:
        yield from line_model_data_list
        yield NewLineExtractInstruction()


def get_default_note_type_for_label(label: str) -> str:
    return label.strip('<>')


def is_parent_path_of(
    parent_path: Sequence[str],
    child_path: Sequence[str]
) -> bool:
    if len(parent_path) >= len(child_path):
        return False
    return tuple(child_path[:len(parent_path)]) == tuple(parent_path)


def is_same_or_parent_path_of(
    parent_path: Sequence[str],
    child_path: Sequence[str]
) -> bool:
    return (
        tuple(parent_path) == tuple(child_path)
        or is_parent_path_of(parent_path, child_path)
    )


class TeiTrainingDataGenerator(ABC):
    @abstractmethod
    def get_training_tei_xml_for_multiple_model_data_iterables(
        self,
        model_data_iterables: Iterable[Iterable[LayoutModelData]]
    ) -> etree.ElementBase:
        pass

    @abstractmethod
    def get_training_tei_xml_for_model_data_iterable(
        self,
        model_data_iterable: Iterable[LayoutModelData]
    ) -> etree.ElementBase:
        pass

    @abstractmethod
    def get_default_tei_filename_suffix(self) -> Optional[str]:
        pass

    def get_default_data_filename_suffix(self) -> Optional[str]:
        return None

    def get_default_tei_sub_directory(self) -> Optional[str]:
        pass

    def get_default_data_sub_directory(self) -> Optional[str]:
        pass


class AbstractTeiTrainingDataGenerator(TeiTrainingDataGenerator):
    def __init__(
        self,
        root_training_xml_element_path: Sequence[str],
        training_xml_element_path_by_label: Mapping[str, Sequence[str]],
        root_tag: str = 'tei',
        use_tei_namespace: bool = True,
        element_maker: Optional[ElementMaker] = None,
        reset_training_xml_element_path_by_label: Optional[Mapping[str, Sequence[str]]] = None,
        default_tei_filename_suffix: Optional[str] = None,
        default_data_filename_suffix: Optional[str] = None,
        default_tei_sub_directory: Optional[str] = None,
        default_data_sub_directory: Optional[str] = None
    ):
        self.root_training_xml_element_path = root_training_xml_element_path
        self.root_parent_training_xml_element_path = root_training_xml_element_path[:-1]
        self.training_xml_element_path_by_label = training_xml_element_path_by_label
        self.reset_training_xml_element_path_by_label = (
            reset_training_xml_element_path_by_label or {}
        )
        self._training_xml_element_paths = {
            tuple(element_path)
            for label, element_path in training_xml_element_path_by_label.items()
            if (
                label not in OTHER_LABELS
                and tuple(element_path) != tuple(root_training_xml_element_path)
            )
        }
        self.other_element_path = training_xml_element_path_by_label.get('<other>')
        if element_maker is None:
            element_maker = TEI_E if use_tei_namespace else NO_NS_TEI_E
        self.element_maker = element_maker
        self.root_tag = root_tag
        self.default_tei_filename_suffix = default_tei_filename_suffix
        self.default_data_filename_suffix = default_data_filename_suffix
        self.default_tei_sub_directory = default_tei_sub_directory
        self.default_data_sub_directory = default_data_sub_directory

    def get_default_tei_filename_suffix(self) -> Optional[str]:
        return self.default_tei_filename_suffix

    def get_default_data_filename_suffix(self) -> Optional[str]:
        return self.default_data_filename_suffix

    def get_default_tei_sub_directory(self) -> Optional[str]:
        return self.default_tei_sub_directory

    def get_default_data_sub_directory(self) -> Optional[str]:
        return self.default_data_sub_directory

    def get_training_xml_path_for_label(
        self,
        label: Optional[str],
        current_path: Sequence[str]
    ) -> Sequence[str]:
        if not label or label in OTHER_LABELS:
            if label and self.other_element_path is not None:
                return self.other_element_path
            if tuple(current_path) in self._training_xml_element_paths:
                LOGGER.debug(
                    'found current path in element paths, returning parent: %r', current_path
                )
                return current_path[:-1]
            LOGGER.debug(
                'not found current path in element paths, returning current: %r', current_path
            )
            return current_path
        training_xml_path = self.training_xml_element_path_by_label.get(label or '')
        if not training_xml_path:
            note_type = get_default_note_type_for_label(label)
            LOGGER.info('label not mapped, creating note: %r', label)
            training_xml_path = (
                list(self.root_training_xml_element_path) + [f'note[@type="{note_type}"]']
            )
        return training_xml_path

    def get_reset_training_xml_path_for_label(
        self,
        label: Optional[str],
        prefix: Optional[str]
    ) -> Optional[Sequence[str]]:
        if prefix != 'B' or not label:
            return None
        return self.reset_training_xml_element_path_by_label.get(label)

    def write_xml_for_model_data_with_instructions_iterable(
        self,
        xml_writer: XmlTreeWriter,
        model_data_or_instruction_iterable: Iterable[Union[LayoutModelData, ExtractInstruction]]
    ):
        default_path = xml_writer.current_path
        LOGGER.debug('default_path: %r', default_path)
        pending_whitespace = ''
        prev_label: str = ''
        pending_reset_path: Optional[List[str]] = None
        for model_data_or_instruction in model_data_or_instruction_iterable:
            if isinstance(model_data_or_instruction, LayoutModelData):
                model_data = model_data_or_instruction
                layout_token = model_data.layout_token
                assert layout_token is not None
                prefixed_label = get_model_data_label(model_data)
                prefix, label = get_split_prefix_label(prefixed_label or '')
                xml_element_path = self.get_training_xml_path_for_label(
                    label,
                    current_path=xml_writer.current_path
                )
                reset_path = self.get_reset_training_xml_path_for_label(
                    label=label,
                    prefix=prefix
                )
                if pending_reset_path is not None:
                    reset_path = pending_reset_path
                    pending_reset_path = None
                LOGGER.debug(
                    'label: %r (%r: %r; reset_path=%r)',
                    label, prefix, xml_element_path, reset_path
                )
                if reset_path is not None:
                    xml_writer.require_path(reset_path)
                elif (
                    prev_label not in OTHER_LABELS
                    and pending_whitespace
                    and not is_same_or_parent_path_of(xml_writer.current_path, xml_element_path)
                ):
                    LOGGER.debug(
                        'closing element before adding whitespace, %r -> %r',
                        xml_writer.current_path, xml_element_path
                    )
                    xml_writer.require_path(xml_writer.current_path[:-1])
                elif prefix == 'B' and label not in OTHER_LABELS:
                    xml_writer.require_path(xml_element_path[:-1])
                xml_writer.require_path_or_below(xml_element_path)
                xml_writer.append_text(pending_whitespace)
                pending_whitespace = ''
                xml_writer.require_path(xml_element_path)
                xml_writer.append_text(layout_token.text)
                pending_whitespace = layout_token.whitespace
                prev_label = label
            elif isinstance(model_data_or_instruction, ResetExtractInstruction):
                pending_reset_path = model_data_or_instruction.reset_element_path
            elif isinstance(model_data_or_instruction, NewLineExtractInstruction):
                xml_writer.append(self.element_maker('lb'))
                pending_whitespace = '\n'
        xml_writer.require_path(default_path)
        xml_writer.append_text(pending_whitespace)

    def iter_model_data_or_instruction_for_model_data_iterable(
        self,
        model_data_iterable: Iterable[LayoutModelData]
    ) -> Iterable[Union[LayoutModelData, ExtractInstruction]]:
        return iter_model_data_with_new_line_instruction(
            model_data_iterable
        )

    def write_xml_for_model_data_iterable(
        self,
        xml_writer: XmlTreeWriter,
        model_data_iterable: Iterable[LayoutModelData]
    ):
        self.write_xml_for_model_data_with_instructions_iterable(
            xml_writer,
            self.iter_model_data_or_instruction_for_model_data_iterable(
                model_data_iterable
            )
        )

    def _get_xml_writer(self) -> XmlTreeWriter:
        return XmlTreeWriter(
            self.element_maker(self.root_tag),
            element_maker=self.element_maker
        )

    def get_post_processed_xml_root(self, xml_root: etree.ElementBase):
        return xml_root

    def get_training_tei_xml_for_multiple_model_data_iterables(
        self,
        model_data_iterables: Iterable[Iterable[LayoutModelData]]
    ) -> etree.ElementBase:
        xml_writer = self._get_xml_writer()
        xml_writer.require_path(self.root_parent_training_xml_element_path)
        for model_data_iterable in model_data_iterables:
            xml_writer.require_path(self.root_parent_training_xml_element_path)
            xml_writer.require_path(self.root_training_xml_element_path)
            self.write_xml_for_model_data_iterable(
                xml_writer,
                model_data_iterable=model_data_iterable
            )
        return self.get_post_processed_xml_root(xml_writer.root)

    def get_training_tei_xml_for_model_data_iterable(
        self,
        model_data_iterable: Iterable[LayoutModelData]
    ) -> etree.ElementBase:
        return self.get_training_tei_xml_for_multiple_model_data_iterables(
            [model_data_iterable]
        )


TEI_LB = 'lb'


LINE_BREAK_TAGS = {
    TEI_LB,
    TEI_NS_PREFIX + TEI_LB
}


def _get_tag_expression_for_element(element: etree.ElementBase) -> str:
    if not element.attrib:
        return element.tag
    if len(element.attrib) > 1:
        raise ValueError('only supporting up to one attribute')
    key, value = list(element.attrib.items())[0]
    return '{tag}[@{key}="{value}"]'.format(tag=element.tag, key=key, value=value)


class TeiTrainingElementPath(NamedTuple):
    element_list: Sequence[etree.ElementBase] = tuple([])

    def get_path(self) -> Sequence[str]:
        return [
            _get_tag_expression_for_element(element)
            for element in self.element_list
        ]

    def append(self, element: etree.ElementBase) -> 'TeiTrainingElementPath':
        return TeiTrainingElementPath(
            list(self.element_list) + [element]
        )


EMPTY_TEI_TRAINING_ELEMENT_PATH = TeiTrainingElementPath()


class TeiTrainingText(NamedTuple):
    text: str
    path: TeiTrainingElementPath
    is_start: bool


class TeiTrainingLine(NamedTuple):
    text_list: Sequence[TeiTrainingText]


def is_line_break_element(element: etree.ElementBase) -> bool:
    return element.tag in LINE_BREAK_TAGS


def _iter_flat_tei_training_text_from_element(
    parent_element: etree.ElementBase,
    current_path: TeiTrainingElementPath = EMPTY_TEI_TRAINING_ELEMENT_PATH
) -> Iterable[Union[TeiTrainingText, ExtractInstruction]]:
    LOGGER.debug('current_path: %s', current_path)
    is_start = True
    if parent_element.text:
        yield TeiTrainingText(
            text=parent_element.text,
            path=current_path,
            is_start=is_start
        )
        is_start = False

    for child_element in parent_element:
        if is_line_break_element(child_element):
            yield NewLineExtractInstruction()
        else:
            child_path = current_path.append(child_element)
            yield from _iter_flat_tei_training_text_from_element(
                child_element,
                child_path
            )

        if child_element.tail:
            yield TeiTrainingText(
                text=child_element.tail,
                path=current_path,
                is_start=is_start
            )
            is_start = False


def _iter_tei_training_lines_from_element(
    parent_element: etree.ElementBase,
    current_path: TeiTrainingElementPath = EMPTY_TEI_TRAINING_ELEMENT_PATH
) -> Iterable[TeiTrainingLine]:
    line_text_list = []
    for item in _iter_flat_tei_training_text_from_element(
        parent_element,
        current_path
    ):
        if isinstance(item, TeiTrainingText):
            line_text_list.append(item)
        elif isinstance(item, NewLineExtractInstruction):
            yield TeiTrainingLine(line_text_list)
            line_text_list = []
        else:
            raise RuntimeError('unrecognised item: %r' % item)
    if line_text_list:
        yield TeiTrainingLine(line_text_list)


T = TypeVar('T')


def iter_group_doc_items_with_new_doc_marker(
    flat_item_iterable: Iterable[Union[T, NewDocumentMarker]]
) -> Iterable[List[T]]:
    doc_items: List[T] = []
    for item in flat_item_iterable:
        if isinstance(item, NewDocumentMarker):
            yield doc_items
            doc_items = []
            continue
        doc_items.append(item)


def iter_tag_result_for_flat_tag_result(
    flat_tag_result_iterable: Iterable[Union[Tuple[str, str], NewDocumentMarker]]
) -> Iterable[List[Tuple[str, str]]]:
    doc_tag_result: List[Tuple[str, str]] = []
    for token_tag_result in flat_tag_result_iterable:
        if isinstance(token_tag_result, NewDocumentMarker):
            yield doc_tag_result
            doc_tag_result = []
            continue
        doc_tag_result.append(token_tag_result)


def get_tag_result_for_flat_tag_result(
    flat_tag_result_iterable: Iterable[Union[Tuple[str, str], NewDocumentMarker]]
) -> List[List[Tuple[str, str]]]:
    return list(iter_tag_result_for_flat_tag_result(flat_tag_result_iterable))


class TrainingTeiParser(ABC):
    @abstractmethod
    def parse_training_tei_to_tag_result(
        self,
        tei_root: etree.ElementBase
    ) -> List[List[Tuple[str, str]]]:
        pass

    @abstractmethod
    def parse_training_tei_to_labeled_layout_tokens_list(
        self,
        tei_root: etree.ElementBase
    ) -> Sequence[Sequence[LabeledLayoutToken]]:
        pass


def get_element_path_with_prefix(
    element_path: Sequence[str],
    prefix: str
) -> Sequence[str]:
    return [
        prefix + item
        for item in element_path
    ]


class AbstractTrainingTeiParser(TrainingTeiParser):
    def __init__(
        self,
        root_training_xml_element_path: Sequence[str],
        training_xml_element_path_by_label: Mapping[str, Sequence[str]],
        use_tei_namespace: bool,
        line_as_token: bool = False,
    ) -> None:
        tag_namespace_prefix = TEI_NS_PREFIX if use_tei_namespace else ''
        if use_tei_namespace:
            root_training_xml_element_path = get_element_path_with_prefix(
                root_training_xml_element_path,
                'tei:'
            )
        self.label_by_relative_element_path_map = {
            tuple(
                get_element_path_with_prefix(
                    element_path[len(root_training_xml_element_path):],
                    tag_namespace_prefix
                )
            ): label
            for label, element_path in training_xml_element_path_by_label.items()
        }
        for element_path in list(self.label_by_relative_element_path_map.keys()):
            if len(element_path) < 2:
                continue
            parent_element_path = element_path[:-1]
            if parent_element_path not in self.label_by_relative_element_path_map:
                self.label_by_relative_element_path_map[parent_element_path] = 'O'
        self.root_training_xml_xpath = './' + '/'.join(root_training_xml_element_path)
        self.line_as_token = line_as_token

    def _get_label_for_element_path(
        self,
        tei_training_element_path: TeiTrainingElementPath,
        text: str
    ) -> str:
        element_path = tei_training_element_path.get_path()
        label = self.label_by_relative_element_path_map.get(tuple(element_path))
        if not label:
            raise RuntimeError(
                'label not found for %r (available: %r; for text: %r)' % (
                    element_path,
                    self.label_by_relative_element_path_map.keys(),
                    text
                )
            )
        return label

    def iter_parse_training_tei_to_flat_labeled_layout_tokens(
        self,
        tei_root: etree.ElementBase
    ) -> Iterable[Union[LabeledLayoutToken, NewDocumentMarker]]:
        for text_node in tei_xpath(tei_root, self.root_training_xml_xpath):
            tei_training_lines = list(
                _iter_tei_training_lines_from_element(
                    text_node, EMPTY_TEI_TRAINING_ELEMENT_PATH
                )
            )
            LOGGER.debug('tei_training_lines: %r', tei_training_lines)
            prefix = ''
            prev_label = ''
            for line_index, line in enumerate(tei_training_lines):
                line_meta = LayoutLineMeta(line_id=1 + line_index)
                for text in line.text_list:
                    if text.text.isspace():
                        continue
                    token_count = 0
                    if text.path.element_list:
                        label = self._get_label_for_element_path(text.path, text=text.text)
                        if prev_label != label:
                            prefix = 'B-' if text.is_start else 'I-'
                    else:
                        label = 'O'
                        prefix = ''
                    if label in OTHER_LABELS:
                        prefix = ''
                    prev_label = label
                    for token_text in get_tokenized_tokens(text.text):
                        yield LabeledLayoutToken(
                            label=prefix + label,
                            layout_token=LayoutToken(
                                text=token_text,
                                line_meta=line_meta
                            )
                        )
                        token_count += 1
                        if prefix:
                            prefix = 'I-'
                        if self.line_as_token:
                            break
                    if token_count and self.line_as_token:
                        # we are only outputting the first token of each line
                        break
            yield NEW_DOCUMENT_MARKER

    def iter_parse_training_tei_to_flat_tag_result(
        self,
        tei_root: etree.ElementBase
    ) -> Iterable[Union[Tuple[str, str], NewDocumentMarker]]:
        for item in self.iter_parse_training_tei_to_flat_labeled_layout_tokens(
            tei_root
        ):
            if isinstance(item, NewDocumentMarker):
                yield item
                continue
            assert isinstance(item, LabeledLayoutToken)
            yield item.layout_token.text, item.label

    def parse_training_tei_to_tag_result(
        self,
        tei_root: etree.ElementBase
    ) -> List[List[Tuple[str, str]]]:
        return list(iter_group_doc_items_with_new_doc_marker(
            self.iter_parse_training_tei_to_flat_tag_result(
                tei_root
            )
        ))

    def parse_training_tei_to_labeled_layout_tokens_list(
        self,
        tei_root: etree.ElementBase
    ) -> Sequence[Sequence[LabeledLayoutToken]]:
        return list(iter_group_doc_items_with_new_doc_marker(
            self.iter_parse_training_tei_to_flat_labeled_layout_tokens(
                tei_root
            )
        ))



from .config import (  # noqa  pylint: disable=unused-import
    FullTextProcessorConfig
)
from .models import (  # noqa  pylint: disable=unused-import
    load_models
)
from .processor import (  # noqa  pylint: disable=unused-import
    FullTextProcessorDocumentContext,
    FullTextProcessor
)



import logging
import threading
from typing import Optional, Union

import PIL.Image

import tesserocr
from tesserocr import PyTessBaseAPI

from sciencebeam_parser.ocr_models.ocr_model import (
    OpticalCharacterRecognitionModel,
    OpticalCharacterRecognitionModelResult,
    SimpleOpticalCharacterRecognitionModelResult
)
from sciencebeam_parser.utils.lazy import LazyLoaded


LOGGER = logging.getLogger(__name__)


DEFAULT_OCR_LANG = 'eng'


def get_enum_value(enum_class, value: Optional[Union[int, str]], default_value: int) -> int:
    if value is None:
        return default_value
    if isinstance(value, int):
        return value
    return getattr(enum_class, value)


class TesserComputerVisionModel(OpticalCharacterRecognitionModel):
    def __init__(self, config: dict):
        super().__init__()
        self._lock = threading.Lock()
        self.lang = str(config.get('lang') or DEFAULT_OCR_LANG)
        self.oem = get_enum_value(tesserocr.OEM, config.get('oem'), tesserocr.OEM.DEFAULT)
        self.psm = get_enum_value(tesserocr.PSM, config.get('psm'), tesserocr.PSM.AUTO)
        self._lazy_tesser_api = LazyLoaded[PyTessBaseAPI](self._create_tesser_api)

    def _create_tesser_api(self) -> PyTessBaseAPI:
        LOGGER.info(
            'creating tesser api with oem=%r, psm=%r, lang=%r',
            self.oem, self.psm, self.lang
        )
        return PyTessBaseAPI(
            oem=self.oem,
            psm=self.psm,
            lang=self.lang
        ).__enter__()

    @property
    def tesser_api(self) -> PyTessBaseAPI:
        return self._lazy_tesser_api.get()

    def preload(self):
        with self._lock:
            self._lazy_tesser_api.get()

    def predict_single(self, image: PIL.Image.Image) -> OpticalCharacterRecognitionModelResult:
        with self._lock:
            tesser_api = self.tesser_api
            LOGGER.info(
                'setting ocr image: %dx%d (format=%r)',
                image.width, image.height, image.format
            )
            tesser_api.SetImage(image)
            text = self.tesser_api.GetUTF8Text()
            return SimpleOpticalCharacterRecognitionModelResult(
                text=text
            )



import logging
from typing import Iterable, Tuple

from sciencebeam_parser.document.layout_document import (
    LayoutBlock
)
from sciencebeam_parser.document.semantic_document import (
    SemanticSection,
    SemanticSectionTypes
)
from sciencebeam_parser.models.fulltext.training_data import (
    FullTextTeiTrainingDataGenerator,
    FullTextTrainingTeiParser
)
from sciencebeam_parser.models.model import Model

from sciencebeam_parser.models.data import (
    DocumentFeaturesContext
)
from sciencebeam_parser.models.fulltext.data import FullTextDataGenerator
from sciencebeam_parser.models.fulltext.extract import FullTextSemanticExtractor


LOGGER = logging.getLogger(__name__)


class FullTextModel(Model):
    def get_data_generator(
        self,
        document_features_context: DocumentFeaturesContext
    ) -> FullTextDataGenerator:
        return FullTextDataGenerator(
            document_features_context=document_features_context
        )

    def get_semantic_extractor(self) -> FullTextSemanticExtractor:
        return FullTextSemanticExtractor()

    def get_tei_training_data_generator(self) -> FullTextTeiTrainingDataGenerator:
        return FullTextTeiTrainingDataGenerator()

    def get_training_tei_parser(self) -> FullTextTrainingTeiParser:
        return FullTextTrainingTeiParser()

    def update_section_with_entity_blocks(
        self,
        parent_section: SemanticSection,
        entity_tokens: Iterable[Tuple[str, LayoutBlock]],
        section_type: str = SemanticSectionTypes.OTHER
    ):
        semantic_extractor = self.get_semantic_extractor()
        for semantic_content in semantic_extractor.iter_semantic_content_for_entity_blocks(
            entity_tokens=entity_tokens,
            section_type=section_type
        ):
            parent_section.add_content(semantic_content)

    def get_section_for_entity_blocks(
        self,
        entity_tokens: Iterable[Tuple[str, LayoutBlock]]
    ) -> SemanticSection:
        parent_section = SemanticSection()
        self.update_section_with_entity_blocks(parent_section, entity_tokens)
        return parent_section



from typing import Tuple


def strip_tag_prefix(tag: str) -> str:
    if tag and (tag.startswith('B-') or tag.startswith('I-')):
        return tag[2:]
    return tag


def get_split_prefix_label(prefixed_tag: str) -> Tuple[str, str]:
    if '-' in prefixed_tag:
        prefix, tag = prefixed_tag.split('-', maxsplit=1)
    else:
        prefix = ''
        tag = prefixed_tag
    return prefix, tag



from typing import Iterable

from sciencebeam_parser.models.data import (
    ContextAwareLayoutTokenFeatures,
    ContextAwareLayoutTokenModelDataGenerator,
    LayoutModelData
)


class CitationDataGenerator(ContextAwareLayoutTokenModelDataGenerator):
    def iter_model_data_for_context_layout_token_features(
        self,
        token_features: ContextAwareLayoutTokenFeatures
    ) -> Iterable[LayoutModelData]:
        yield token_features.get_layout_model_data([
            token_features.token_text,
            token_features.get_lower_token_text(),
            token_features.get_prefix(1),
            token_features.get_prefix(2),
            token_features.get_prefix(3),
            token_features.get_prefix(4),
            token_features.get_suffix(1),
            token_features.get_suffix(2),
            token_features.get_suffix(3),
            token_features.get_suffix(4),
            token_features.get_line_status_with_lineend_for_single_token(),
            token_features.get_capitalisation_status_using_allcap(),
            token_features.get_digit_status_using_containsdigits(),
            token_features.get_str_is_single_char(),
            token_features.get_dummy_str_is_proper_name(),
            token_features.get_dummy_str_is_common_name(),
            token_features.get_str_is_first_name(),
            token_features.get_str_is_last_name(),
            token_features.get_dummy_str_is_location_name(),
            token_features.get_dummy_str_is_year(),
            token_features.get_dummy_str_is_month(),
            token_features.get_dummy_str_is_http(),
            token_features.get_dummy_str_is_known_collaboration(),
            token_features.get_dummy_str_is_known_journal_title(),
            token_features.get_dummy_str_is_known_conference_title(),
            token_features.get_dummy_str_is_known_publisher(),
            token_features.get_dummy_str_is_known_identifier(),
            token_features.get_punctuation_type_feature(),
            token_features.get_str_sentence_token_relative_position(),
            token_features.get_dummy_label()
        ])



import logging
import math
import re
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Iterable, List, NamedTuple, Optional

from lxml import etree

from sciencebeam_parser.lookup import TextLookUp
from sciencebeam_parser.document.layout_document import LayoutDocument, LayoutToken,  LayoutLine
from sciencebeam_parser.external.pdfalto.parser import parse_alto_root


LOGGER = logging.getLogger(__name__)


class AppFeaturesContext(NamedTuple):
    country_lookup: Optional[TextLookUp] = None
    first_name_lookup: Optional[TextLookUp] = None
    last_name_lookup: Optional[TextLookUp] = None


DEFAULT_APP_FEATURES_CONTEXT = AppFeaturesContext()


class DocumentFeaturesContext(NamedTuple):
    app_features_context: AppFeaturesContext = DEFAULT_APP_FEATURES_CONTEXT


DEFAULT_DOCUMENT_FEATURES_CONTEXT = DocumentFeaturesContext()


class NewDocumentMarker:
    pass


NEW_DOCUMENT_MARKER = NewDocumentMarker()


class LabeledLayoutToken(NamedTuple):
    label: str
    layout_token: LayoutToken


@dataclass
class LayoutModelData:
    data_line: str
    layout_line: Optional[LayoutLine] = field(repr=False, default=None)
    layout_token: Optional[LayoutToken] = field(repr=False, default=None)

    @property
    def label_token_text(self):
        return self.data_line.split(' ', maxsplit=1)[0]


@dataclass
class LabeledLayoutModelData(LayoutModelData):
    label: Optional[str] = None

    @staticmethod
    def from_model_data(
        model_data: LayoutModelData,
        label: Optional[str] = None
    ) -> 'LabeledLayoutModelData':
        return LabeledLayoutModelData(
            data_line=model_data.data_line,
            layout_line=model_data.layout_line,
            layout_token=model_data.layout_token,
            label=label
        )


class ModelDataGenerator(ABC):
    def iter_data_lines_for_xml_root(
        self,
        root: etree.ElementBase
    ) -> Iterable[str]:
        return self.iter_data_lines_for_layout_document(
            parse_alto_root(root)
        )

    @abstractmethod
    def iter_model_data_for_layout_document(
        self,
        layout_document: LayoutDocument
    ) -> Iterable[LayoutModelData]:
        pass

    def iter_data_lines_for_layout_document(  # pylint: disable=too-many-locals
        self,
        layout_document: LayoutDocument
    ) -> Iterable[str]:
        return (
            model_data.data_line
            for model_data in self.iter_model_data_for_layout_document(
                layout_document
            )
        )

    def iter_data_lines_for_layout_documents(  # pylint: disable=too-many-locals
        self,
        layout_documents: Iterable[LayoutDocument]
    ) -> Iterable[str]:
        for index, layout_document in enumerate(layout_documents):
            LOGGER.debug('generating data lines for document: index=%d', index)
            if index > 0:
                LOGGER.debug('adding document separator')
                yield from ['\n']
            yield from (
                model_data.data_line
                for model_data in self.iter_model_data_for_layout_document(
                    layout_document
                )
            )


def feature_linear_scaling_int(pos: int, total: int, bin_count: int) -> int:
    """
    Given an integer value between 0 and total, discretized into nbBins following a linear scale
    Adapted from:
    grobid-core/src/main/java/org/grobid/core/features/FeatureFactory.java
    """
    if pos >= total:
        return bin_count
    if pos <= 0:
        return 0
    return math.floor((pos / total) * bin_count)


def get_token_font_status(previous_token: Optional[LayoutToken], current_token: LayoutToken):
    if not previous_token:
        return 'NEWFONT'
    return (
        'SAMEFONT' if current_token.font.font_family == previous_token.font.font_family
        else 'NEWFONT'
    )


def get_token_font_size_feature(
    previous_token: Optional[LayoutToken],
    current_token: LayoutToken
):
    if not previous_token:
        return 'HIGHERFONT'
    previous_font_size = previous_token.font.font_size
    current_font_size = current_token.font.font_size
    if not previous_font_size or not current_font_size:
        return 'HIGHERFONT'
    if previous_font_size < current_font_size:
        return 'HIGHERFONT'
    if previous_font_size > current_font_size:
        return 'LOWERFONT'
    return 'SAMEFONTSIZE'


def get_digit_feature(text: str) -> str:
    if text.isdigit():
        return 'ALLDIGIT'
    for c in text:
        if c.isdigit():
            return 'CONTAINSDIGITS'
    return 'NODIGIT'


def get_capitalisation_feature(text: str) -> str:
    if text and all(not c.islower() for c in text):
        return 'ALLCAP'
    if text and text[0].isupper():
        return 'INITCAP'
    return 'NOCAPS'


class PunctuationProfileValues:
    OPENBRACKET = 'OPENBRACKET'
    ENDBRACKET = 'ENDBRACKET'
    DOT = 'DOT'
    COMMA = 'COMMA'
    HYPHEN = 'HYPHEN'
    QUOTE = 'QUOTE'
    PUNCT = 'PUNCT'
    NOPUNCT = 'NOPUNCT'


PUNCTUATION_PROFILE_MAP = {
    '(': PunctuationProfileValues.OPENBRACKET,
    '[': PunctuationProfileValues.OPENBRACKET,
    ')': PunctuationProfileValues.ENDBRACKET,
    ']': PunctuationProfileValues.ENDBRACKET,
    '.': PunctuationProfileValues.DOT,
    ',': PunctuationProfileValues.COMMA,
    '-': PunctuationProfileValues.HYPHEN,
    '–': PunctuationProfileValues.HYPHEN,
    '"': PunctuationProfileValues.QUOTE,
    '\'': PunctuationProfileValues.QUOTE,
    '`': PunctuationProfileValues.QUOTE,
    '’': PunctuationProfileValues.QUOTE
}


IS_PUNCT_PATTERN = r"^[\,\:;\?\.]+$"


PUNCTUATION_PROFILE_CHARACTERS = (
    "(（[ •*,:;?.!/)）-−–‐«»„\"“”‘’'`$#@]*\u2666\u2665\u2663\u2660\u00A0"
)


def get_line_status_with_lineend_for_single_token(token_index: int, token_count: int) -> str:
    return (
        'LINEEND' if token_index == token_count - 1
        else (
            'LINESTART' if token_index == 0
            else 'LINEIN'
        )
    )


def get_line_status_with_linestart_for_single_token(
    token_index: int, token_count: int
) -> str:
    return (
        'LINESTART' if token_index == 0
        else (
            'LINEEND' if token_index == token_count - 1
            else 'LINEIN'
        )
    )


def get_block_status_with_blockend_for_single_token(
    line_index: int,
    line_count: int,
    line_status: str
) -> str:
    return (
        'BLOCKEND'
        if line_index == line_count - 1 and line_status == 'LINEEND'
        else (
            'BLOCKSTART'
            if line_index == 0 and line_status == 'LINESTART'
            else 'BLOCKIN'
        )
    )


def get_block_status_with_blockstart_for_single_token(
    line_index: int,
    line_count: int,
    line_status: str
) -> str:
    return (
        'BLOCKSTART'
        if line_index == 0 and line_status == 'LINESTART'
        else (
            'BLOCKEND'
            if line_index == line_count - 1 and line_status == 'LINEEND'
            else 'BLOCKIN'
        )
    )


class RelativeFontSizeFeature:
    def __init__(self, layout_tokens: Iterable[LayoutToken]):
        font_sizes = [
            layout_token.font.font_size
            for layout_token in layout_tokens
            if layout_token.font.font_size
        ]
        LOGGER.debug('font_sizes (%d): %r', len(font_sizes), font_sizes)
        self.largest_font_size = max(font_sizes) if font_sizes else 0.0
        self.smallest_font_size = min(font_sizes) if font_sizes else 0.0
        self.mean_font_size = sum(font_sizes) / len(font_sizes) if font_sizes else 0.0
        LOGGER.debug('relative font size: %r', self)

    def __repr__(self) -> str:
        return (
            '%s(largest_font_size=%f, smallest_font_size=%f, mean_font_size=%f)'
        ) % (
            type(self).__name__,
            self.largest_font_size,
            self.smallest_font_size,
            self.mean_font_size
        )

    def is_largest_font_size(self, layout_token: LayoutToken):
        return layout_token.font.font_size == self.largest_font_size

    def is_smallest_font_size(self, layout_token: LayoutToken):
        return layout_token.font.font_size == self.smallest_font_size

    def is_larger_than_average_font_size(self, layout_token: LayoutToken):
        if not layout_token.font.font_size:
            return False
        return layout_token.font.font_size > self.mean_font_size


class LineIndentationStatusFeature:
    def __init__(self):
        self._line_start_x = None
        self._is_new_line = True
        self._is_indented = False

    def on_new_block(self):
        pass

    def on_new_line(self):
        self._is_new_line = True

    def get_is_indented_and_update(self, layout_token: LayoutToken):
        if self._is_new_line and layout_token.coordinates and layout_token.text:
            previous_line_start_x = self._line_start_x
            self._line_start_x = layout_token.coordinates.x
            character_width = layout_token.coordinates.width / len(layout_token.text)
            if previous_line_start_x is not None:
                if self._line_start_x - previous_line_start_x > character_width:
                    self._is_indented = True
                if previous_line_start_x - self._line_start_x > character_width:
                    self._is_indented = False
        self._is_new_line = False
        return self._is_indented


def get_punctuation_type_feature(text: str) -> str:
    result = PUNCTUATION_PROFILE_MAP.get(text)
    if not result and re.match(IS_PUNCT_PATTERN, text):
        return PunctuationProfileValues.PUNCT
    if not result:
        return PunctuationProfileValues.NOPUNCT
    return result


def get_raw_punctuation_profile_feature(text: str) -> str:
    if not text:
        return ''
    return ''.join((
        c
        for c in text
        if not c.isspace() and c in PUNCTUATION_PROFILE_CHARACTERS
    ))


def get_punctuation_profile_feature_for_raw_punctuation_profile_feature(
    raw_punctuation_profile: str
) -> str:
    if not raw_punctuation_profile:
        return 'no'
    return raw_punctuation_profile


def get_punctuation_profile_length_for_raw_punctuation_profile_feature(
    raw_punctuation_profile: str,
    max_length: Optional[int] = None
) -> str:
    if max_length:
        return str(min(10, len(raw_punctuation_profile)))
    return str(len(raw_punctuation_profile))


def get_char_shape_feature(ch: str) -> str:
    if ch.isdigit():
        return 'd'
    if ch.isalpha():
        if ch.isupper():
            return 'X'
        return 'x'
    return ch


def get_word_shape_feature(text: str) -> str:
    shape = [
        get_char_shape_feature(ch)
        for ch in text
    ]
    prefix = shape[:1]
    middle = shape[1:-2]
    suffix = shape[1:][-2:]
    middle_without_consequitive_duplicates = middle[:1].copy()
    for ch in middle[1:]:
        if ch != middle_without_consequitive_duplicates[-1]:
            middle_without_consequitive_duplicates.append(ch)
    return ''.join(prefix + middle_without_consequitive_duplicates + suffix)


def get_str_bool_feature_value(value: Optional[bool]) -> str:
    return '1' if value else '0'


class CommonLayoutTokenFeatures(ABC):  # pylint: disable=too-many-public-methods
    def __init__(self, layout_token: LayoutToken) -> None:
        self.layout_token = layout_token
        self.token_text = layout_token.text or ''

    def get_lower_token_text(self) -> str:
        return self.token_text.lower()

    def get_prefix(self, n: int) -> str:
        return self.token_text[:n]

    def get_suffix(self, n: int) -> str:
        return self.token_text[-n:]

    def get_str_is_bold(self) -> str:
        return get_str_bool_feature_value(self.layout_token.font.is_bold)

    def get_str_is_italic(self) -> str:
        return get_str_bool_feature_value(self.layout_token.font.is_italics)

    def get_str_is_superscript(self) -> str:
        return get_str_bool_feature_value(self.layout_token.font.is_superscript)

    def get_str_is_single_char(self) -> str:
        return get_str_bool_feature_value(len(self.token_text) == 1)

    def get_digit_status_using_containsdigits(self) -> str:
        return get_digit_feature(self.token_text)

    def get_digit_status_using_containdigit(self) -> str:
        digit_status = get_digit_feature(self.token_text)
        if digit_status == 'CONTAINSDIGITS':
            digit_status = 'CONTAINDIGIT'
        return digit_status

    def get_capitalisation_status_using_allcap(self) -> str:
        if self.get_digit_status_using_containsdigits() == 'ALLDIGIT':
            return 'NOCAPS'
        return get_capitalisation_feature(self.token_text)

    def get_capitalisation_status_using_allcaps(self) -> str:
        capitalisation_status = self.get_capitalisation_status_using_allcap()
        if capitalisation_status == 'ALLCAP':
            return 'ALLCAPS'
        return capitalisation_status

    def get_punctuation_type_feature(self) -> str:
        return get_punctuation_type_feature(self.token_text)

    def get_word_shape_feature(self) -> str:
        return get_word_shape_feature(self.token_text)

    def get_dummy_str_is_proper_name(self) -> str:
        return '0'

    def get_dummy_str_is_common_name(self) -> str:
        return '0'

    def get_dummy_str_is_first_name(self) -> str:
        return '0'

    def get_dummy_str_is_last_name(self) -> str:
        return '0'

    def get_dummy_str_is_known_title(self) -> str:
        return '0'

    def get_dummy_str_is_known_suffix(self) -> str:
        return '0'

    def get_dummy_str_is_location_name(self) -> str:
        return '0'

    def get_dummy_str_is_country_name(self) -> str:
        return '0'

    def get_dummy_str_is_year(self) -> str:
        return '0'

    def get_dummy_str_is_month(self) -> str:
        return '0'

    def get_dummy_str_is_email(self) -> str:
        return '0'

    def get_dummy_str_is_http(self) -> str:
        return '0'

    def get_dummy_str_is_known_collaboration(self) -> str:
        return '0'

    def get_dummy_str_is_known_journal_title(self) -> str:
        return '0'

    def get_dummy_str_is_known_conference_title(self) -> str:
        return '0'

    def get_dummy_str_is_known_publisher(self) -> str:
        return '0'

    def get_dummy_str_is_known_identifier(self) -> str:
        return '0'

    def get_dummy_label(self) -> str:
        return '0'


_LINESCALE = 10


class ContextAwareLayoutTokenFeatures(  # pylint: disable=too-many-public-methods
    CommonLayoutTokenFeatures
):
    def __init__(  # pylint: disable=too-many-locals
        self,
        layout_token: LayoutToken,
        layout_line: LayoutLine,
        document_features_context: DocumentFeaturesContext,
        previous_layout_token: Optional[LayoutToken] = None,
        token_index: int = 0,
        token_count: int = 0,
        document_token_index: int = 0,
        document_token_count: int = 0,
        line_index: int = 0,
        line_count: int = 0,
        concatenated_line_tokens_text: str = '',
        max_concatenated_line_tokens_length: int = 0,
        line_token_position: int = 0,
        relative_font_size_feature: Optional[RelativeFontSizeFeature] = None,
        line_indentation_status_feature: Optional[LineIndentationStatusFeature] = None
    ) -> None:
        super().__init__(layout_token)
        self.layout_line = layout_line
        self.previous_layout_token = previous_layout_token
        self.document_features_context = document_features_context
        self.token_index = token_index
        self.token_count = token_count
        self.document_token_index = document_token_index
        self.document_token_count = document_token_count
        self.line_index = line_index
        self.line_count = line_count
        self.concatenated_line_tokens_text = concatenated_line_tokens_text
        self.max_concatenated_line_tokens_length = max_concatenated_line_tokens_length
        self.line_token_position = line_token_position
        self.relative_font_size_feature = relative_font_size_feature
        self.line_indentation_status_feature = line_indentation_status_feature

    def get_layout_model_data(self, features: List[str]) -> LayoutModelData:
        return LayoutModelData(
            layout_line=self.layout_line,
            layout_token=self.layout_token,
            data_line=' '.join(features)
        )

    def get_line_status_with_lineend_for_single_token(self) -> str:
        return get_line_status_with_lineend_for_single_token(
            token_index=self.token_index, token_count=self.token_count
        )

    def get_line_status_with_linestart_for_single_token(self) -> str:
        return get_line_status_with_linestart_for_single_token(
            token_index=self.token_index, token_count=self.token_count
        )

    def get_block_status_with_blockend_for_single_token(self) -> str:
        return get_block_status_with_blockend_for_single_token(
            line_index=self.line_index,
            line_count=self.line_count,
            line_status=self.get_line_status_with_lineend_for_single_token()
        )

    def get_block_status_with_blockstart_for_single_token(self) -> str:
        return get_block_status_with_blockstart_for_single_token(
            line_index=self.line_index,
            line_count=self.line_count,
            line_status=self.get_line_status_with_linestart_for_single_token()
        )

    def get_dummy_page_status(self) -> str:
        return 'PAGEIN'

    def get_is_indented_and_update(self) -> bool:
        assert self.line_indentation_status_feature
        return self.line_indentation_status_feature.get_is_indented_and_update(
            self.layout_token
        )

    def get_alignment_status(self) -> str:
        indented = self.get_is_indented_and_update()
        return 'LINEINDENT' if indented else 'ALIGNEDLEFT'

    def get_token_font_status(self) -> str:
        return get_token_font_status(self.previous_layout_token, self.layout_token)

    def get_token_font_size_feature(self) -> str:
        return get_token_font_size_feature(self.previous_layout_token, self.layout_token)

    def get_str_is_largest_font_size(self) -> str:
        assert self.relative_font_size_feature
        return get_str_bool_feature_value(
            self.relative_font_size_feature.is_largest_font_size(
                self.layout_token
            )
        )

    def get_dummy_str_is_smallest_font_size(self) -> str:
        return '0'

    def get_str_is_smallest_font_size(self) -> str:
        assert self.relative_font_size_feature
        return get_str_bool_feature_value(
            self.relative_font_size_feature.is_smallest_font_size(
                self.layout_token
            )
        )

    def get_dummy_str_is_larger_than_average_font_size(self, value: str = '0') -> str:
        return value

    def get_str_is_larger_than_average_font_size(self) -> str:
        assert self.relative_font_size_feature
        return get_str_bool_feature_value(
            self.relative_font_size_feature.is_larger_than_average_font_size(
                self.layout_token
            )
        )

    def get_raw_line_punctuation_profile(self) -> str:
        return get_raw_punctuation_profile_feature(self.concatenated_line_tokens_text)

    def get_line_punctuation_profile(self) -> str:
        return get_punctuation_profile_feature_for_raw_punctuation_profile_feature(
            self.get_raw_line_punctuation_profile()
        )

    def get_line_punctuation_profile_length_feature(self) -> str:
        return get_punctuation_profile_length_for_raw_punctuation_profile_feature(
            self.get_raw_line_punctuation_profile()
        )

    def get_truncated_line_punctuation_profile_length_feature(self) -> str:
        return get_punctuation_profile_length_for_raw_punctuation_profile_feature(
            self.get_raw_line_punctuation_profile(),
            max_length=10
        )

    def get_str_line_token_relative_position(self) -> str:
        return str(feature_linear_scaling_int(
            self.line_token_position,
            len(self.concatenated_line_tokens_text),
            _LINESCALE
        ))

    def get_str_line_relative_length(self) -> str:
        return str(feature_linear_scaling_int(
            len(self.concatenated_line_tokens_text),
            self.max_concatenated_line_tokens_length,
            _LINESCALE
        ))

    def get_str_sentence_token_relative_position(self) -> str:
        return str(feature_linear_scaling_int(
            # the document is currently the sentence view
            self.document_token_index,
            self.document_token_count,
            12
        ))

    def _get_str_lookup(self, lookup: Optional[TextLookUp]) -> str:
        if not lookup:
            return get_str_bool_feature_value(False)
        return get_str_bool_feature_value(
            lookup.contains(self.token_text)
        )

    def get_str_is_country(self) -> str:
        return self._get_str_lookup(
            self.document_features_context.app_features_context.country_lookup
        )

    def get_str_is_first_name(self) -> str:
        return self._get_str_lookup(
            self.document_features_context.app_features_context.first_name_lookup
        )

    def get_str_is_last_name(self) -> str:
        return self._get_str_lookup(
            self.document_features_context.app_features_context.last_name_lookup
        )

    def get_dummy_str_relative_document_position(self):
        # position within whole document
        return '0'

    def get_dummy_str_relative_page_position(self):
        return '0'

    def get_dummy_str_is_bitmap_around(self) -> str:
        return '0'

    def get_dummy_str_is_vector_around(self) -> str:
        return '0'

    def get_dummy_callout_type(self) -> str:
        return 'UNKNOWN'  # one of UNKNOWN, NUMBER, AUTHOR

    def get_dummy_str_is_callout_known(self) -> str:
        return '0'


class ContextAwareLayoutTokenModelDataGenerator(ModelDataGenerator):
    def __init__(
        self,
        document_features_context: DocumentFeaturesContext
    ):
        self.document_features_context = document_features_context

    @abstractmethod
    def iter_model_data_for_context_layout_token_features(
        self,
        token_features: ContextAwareLayoutTokenFeatures
    ) -> Iterable[LayoutModelData]:
        pass

    def iter_model_data_for_layout_document(  # pylint: disable=too-many-locals
        self,
        layout_document: LayoutDocument
    ) -> Iterable[LayoutModelData]:
        relative_font_size_feature = RelativeFontSizeFeature(
            layout_document.iter_all_tokens()
        )
        line_indentation_status_feature = LineIndentationStatusFeature()
        previous_layout_token: Optional[LayoutToken] = None
        concatenated_line_tokens_length_by_line_id = {
            id(line): sum((len(token.text) for token in line.tokens))
            for block in layout_document.iter_all_blocks()
            for line in block.lines
        }
        if not concatenated_line_tokens_length_by_line_id:
            LOGGER.debug('empty layout document')
            return
        max_concatenated_line_tokens_length = max(
            concatenated_line_tokens_length_by_line_id.values()
        )
        document_token_count = sum((
            1
            for _ in layout_document.iter_all_tokens()
        ))
        document_token_index = 0
        for block in layout_document.iter_all_blocks():
            block_lines = block.lines
            line_count = len(block_lines)
            for line_index, line in enumerate(block_lines):
                line_indentation_status_feature.on_new_line()
                line_tokens = line.tokens
                token_count = len(line_tokens)
                concatenated_line_tokens_text = ''.join([
                    token.text for token in line_tokens
                ])
                line_token_position = 0
                for token_index, token in enumerate(line_tokens):
                    yield from self.iter_model_data_for_context_layout_token_features(
                        ContextAwareLayoutTokenFeatures(
                            token,
                            layout_line=line,
                            previous_layout_token=previous_layout_token,
                            document_features_context=self.document_features_context,
                            token_index=token_index,
                            token_count=token_count,
                            document_token_index=document_token_index,
                            document_token_count=document_token_count,
                            line_index=line_index,
                            line_count=line_count,
                            concatenated_line_tokens_text=concatenated_line_tokens_text,
                            max_concatenated_line_tokens_length=max_concatenated_line_tokens_length,
                            line_token_position=line_token_position,
                            relative_font_size_feature=relative_font_size_feature,
                            line_indentation_status_feature=line_indentation_status_feature
                        )
                    )
                    previous_layout_token = token
                    line_token_position += len(token.text)
                    document_token_index += 1



from lxml import etree

from sciencebeam_parser.document.semantic_document import (
    SemanticContentWrapper,
    SemanticPageRange,
)
from sciencebeam_parser.document.tei.common import (
    TEI_E
)
from sciencebeam_parser.document.tei.factory import (
    SingleElementTeiElementFactory,
    TeiElementFactoryContext
)


class TeiBiblScopeForPageRangeElementFactory(SingleElementTeiElementFactory):
    def get_tei_element_for_semantic_content(
        self,
        semantic_content: SemanticContentWrapper,
        context: TeiElementFactoryContext
    ) -> etree.ElementBase:
        assert isinstance(semantic_content, SemanticPageRange)
        page_range = semantic_content
        if not page_range.from_page or not page_range.to_page:
            return TEI_E(
                'biblScope',
                {'unit': 'page'},
                page_range.get_text()
            )
        return TEI_E(
            'biblScope',
            {
                'unit': 'page',
                'from': page_range.from_page,
                'to': page_range.to_page
            }
        )



from typing import NamedTuple

from typing_extensions import Protocol

from sciencebeam_parser.config.config import AppConfig
from sciencebeam_parser.external.wapiti.wrapper import LazyWapitiBinaryWrapper


# using protocol to avoid delft import where we just need the typing hint
class DownloadManagerProtocol(Protocol):
    def get_local_file(self, file_url: str, auto_uncompress: bool = True) -> str:
        pass

    def is_downloaded(self, file_url: str, auto_uncompress: bool = True) -> bool:
        pass

    def download(
        self, file_url: str,
        local_file: str = None,
        auto_uncompress: bool = True,
        skip_if_downloaded: bool = True
    ) -> str:
        pass

    def download_if_url(self, file_url_or_path: str, **kwargs) -> str:
        pass


class AppContext(NamedTuple):
    app_config: AppConfig
    download_manager: DownloadManagerProtocol
    lazy_wapiti_binary_wrapper: LazyWapitiBinaryWrapper



#!/usr/bin/env python3

# This script will run using the default Python 3 environment
# where LibreOffice's scripts are installed to (at least in Ubuntu).
# This converter is very similar to unoconv but has an option to remove
# line numbers, it is also simpler by being more tailored to the use-case.
# https://github.com/dagwieers/unoconv

from __future__ import absolute_import, print_function

import argparse
import os
import sys
import logging
import subprocess
import atexit
from time import sleep
from contextlib import contextmanager
from typing import Optional, Sequence

# pylint: disable=import-error
import uno  # type: ignore
from com.sun.star.beans import PropertyValue  # type: ignore
from com.sun.star.connection import NoConnectException  # type: ignore
from com.sun.star.document import RedlineDisplayType  # type: ignore
# pylint: enable=import-error


LOGGER = logging.getLogger(__name__)


FILTER_NAME_BY_EXT = {
    'doc': 'MS Word 97',
    'docx': 'Office Open XML Text',
    'dotx': 'Office Open XML Text',
    'rtf': 'Rich Text Format',
    'pdf': 'writer_web_pdf_Export'
}

VALID_OUTPUT_FORMATS = sorted(FILTER_NAME_BY_EXT.keys())


def parse_args(argv: Optional[Sequence[str]] = None):
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(dest='command')

    convert_parser = subparsers.add_parser('convert')
    convert_parser.add_argument(
        '-f', '--format', type=str, required=True,
        choices=VALID_OUTPUT_FORMATS,
        help='Output format (ext)'
    )
    convert_parser.add_argument(
        '-p', '--port', type=int, default=2002,
        help='Port to the uno listener'
    )

    convert_parser.add_argument(
        '--output-file', type=str,
        help='Output file (if specified, only one input file should be used)'
    )

    convert_parser.add_argument(
        'input_file', type=str, nargs='+',
        help='Input files (does not support pdf)'
    )

    convert_parser.add_argument(
        '--remove-line-no', action='store_true', default=False,
        help='remove line number'
    )

    convert_parser.add_argument(
        '--remove-header-footer', action='store_true', default=False,
        help='remove header and footer (including page number)'
    )

    convert_parser.add_argument(
        '--remove-redline', action='store_true', default=False,
        help='remove redlines (track changes, by accepting all changes)'
    )

    convert_parser.add_argument(
        '--keep-listener-running', action='store_true', default=False,
        help='keep listener running in the background'
    )
    convert_parser.add_argument(
        '-n', '--no-launch', action='store_true', default=False,
        help='fail if no listener is found (default: launch one)'
    )

    start_listener_parser = subparsers.add_parser('start-listener')
    start_listener_parser.add_argument(
        '-p', '--port', type=int, default=2002,
        help='Port to the uno listener'
    )

    parser.add_argument(
        '--debug', action='store_true', default=False,
        help='enable debug output'
    )

    args = parser.parse_args(argv)

    if args.debug:
        logging.getLogger().setLevel('DEBUG')

    LOGGER.debug('args: %s', args)

    return args


def get_start_listener_command(port: int) -> Sequence[str]:
    return [
        'soffice',
        '--headless',
        '--invisible',
        '--nocrashreport',
        '--nodefault',
        '--nofirststartwizard',
        '--nologo',
        '--norestore',
        '--accept=socket,host=localhost,port={port};urp;StarOffice.ServiceManager'.format(
            port=port
        )
    ]


def get_resolver():
    local_context = uno.getComponentContext()
    resolver = local_context.ServiceManager.createInstanceWithContext(
        "com.sun.star.bridge.UnoUrlResolver", local_context
    )
    return resolver


def connect(resolver, port: int):
    return resolver.resolve(
        "uno:socket,host=localhost,port={port};urp;StarOffice.ComponentContext".format(
            port=port
        )
    )


def connect_with_timeout(resolver, port: int, timeout: float):
    delay = 0.5
    elapsed = 0.0
    while True:
        try:
            connect_result = connect(resolver, port)
            LOGGER.debug('connected to port %s', port)
            return connect_result
        except NoConnectException as e:
            if elapsed >= timeout:
                LOGGER.debug(
                    'connection failed, timeout exceeded (%.1f >= %s)',
                    elapsed, timeout
                )
                raise e
            LOGGER.debug('connection failed, try again in %.1f (%.1f)', delay, elapsed)
            sleep(delay)
            elapsed += delay


def start_listener(port: int) -> subprocess.Popen:
    LOGGER.debug('starting listener on port %d', port)
    return subprocess.Popen(
        get_start_listener_command(port)
    )


def stop_listener(listener_process: subprocess.Popen):
    LOGGER.debug('stopping listener process with pid: %s', listener_process.pid)
    return listener_process.terminate()


@contextmanager
def managed_connection(resolver, port: int, no_launch: bool, keep_listener_running: bool):
    timeout = 10
    try:
        yield connect_with_timeout(resolver, port, timeout)
    except NoConnectException as e:
        if no_launch:
            raise e
        LOGGER.debug('failed to connect, try to start listener')
        listener_process = start_listener(port)
        try:
            yield connect_with_timeout(resolver, port, timeout)
        finally:
            if not keep_listener_running:
                stop_listener(listener_process)


@contextmanager
def managed_desktop(connection, keep_listener_running: bool):
    LOGGER.debug('starting desktop session')
    desktop = connection.ServiceManager.createInstanceWithContext(
        "com.sun.star.frame.Desktop", connection
    )
    try:
        yield desktop
    finally:
        try:
            if not keep_listener_running:
                LOGGER.debug('terminate desktop session')
                desktop.terminate()
        except Exception as e:  # pylint: disable=broad-except
            LOGGER.warning('caught exception while terminating desktop: %s', e)


def create_property_value(name, value):
    property_value = PropertyValue()
    property_value.Name = name
    property_value.Value = value
    return property_value


def dict_to_property_values(d):
    return tuple((
        create_property_value(key, value)
        for key, value in d.items()
    ))


def property_set_to_dict(property_set):
    return {
        prop.Name: property_set.getPropertyValue(prop.Name)
        for prop in property_set.getPropertySetInfo().getProperties()
    }


def disable_document_header_footer(document):
    styleFamilies = document.getStyleFamilies()
    pageStyles = styleFamilies.getByName('PageStyles')
    if not styleFamilies.hasByName('PageStyles'):
        return
    for styleName in pageStyles.getElementNames():
        pageStyle = pageStyles.getByName(styleName)
        pageStyle.setPropertyValue('HeaderIsOn', False)
        pageStyle.setPropertyValue('FooterIsOn', False)


def convert_document_file(
    desktop,
    input_file: str,
    output_file: str,
    output_ext: str,
    remove_line_no: bool = False,
    remove_redline: bool = False,
    remove_header_footer: bool = False
):
    output_filter_name = FILTER_NAME_BY_EXT[output_ext]

    input_file_url = uno.systemPathToFileUrl(os.path.realpath(input_file))
    document = desktop.loadComponentFromURL(
        input_file_url,
        "_blank", 0,
        dict_to_property_values({'Hidden': True, 'ReadOnly': True})
    )

    if not document:
        raise RuntimeError('failed to load document: %s' % input_file_url)

    try:
        if remove_line_no:
            document.getLineNumberingProperties().IsOn = False

        if remove_header_footer:
            disable_document_header_footer(document)

        if remove_redline:
            document.setPropertyValue('RedlineDisplayType', RedlineDisplayType.NONE)

        output_url = "file://" + os.path.abspath(output_file)
        LOGGER.debug("output_url: %s", output_url)
        document.storeToURL(
            output_url,
            dict_to_property_values({'FilterName': output_filter_name})
        )
    finally:
        # close, parameter: DeliverOwnership
        #    "true: delegates the ownership of ths closing object to any one
        #    which throw the CloseVetoException.
        #    This new owner has to close the closing object again
        #    if his still running processes will be finished."
        document.close(True)


def convert(desktop, args: argparse.Namespace):
    if args.output_file and len(args.input_file) > 1:
        raise RuntimeError(
            ''.join([
                'only one input field should be specified together with --output-file.'
                ' (input files: %s)'
            ]) % args.input_file
        )
    for input_filename in args.input_file:
        LOGGER.info(
            'processing: %s (%s)',
            input_filename,
            '{:,d}'.format(os.path.getsize(input_filename))
        )
        name, input_ext = os.path.splitext(input_filename)
        if input_ext.startswith('.'):
            input_ext = input_ext[1:]
        if not args.output_file and input_ext == args.format:
            raise RuntimeError(
                ''.join([
                    'input and output format should not be the same',
                    ' (unless --output-file was specified): %s -> %s'
                 ]) % (
                    input_ext, args.format
                )
            )
        if args.output_file:
            output_filename = args.output_file
        else:
            output_filename = name + '.' + args.format
        convert_document_file(
            desktop,
            input_filename,
            output_filename,
            args.format,
            remove_line_no=args.remove_line_no,
            remove_header_footer=args.remove_header_footer,
            remove_redline=args.remove_redline
        )


def run(args: argparse.Namespace):
    if args.command == 'convert':
        resolver = get_resolver()
        with managed_connection(
                resolver, args.port,
                no_launch=args.no_launch,
                keep_listener_running=args.keep_listener_running) as connection:

            with managed_desktop(connection, args.keep_listener_running) as desktop:
                convert(desktop, args)
    elif args.command == 'start-listener':
        p = start_listener(args.port)
        atexit.register(
            lambda: stop_listener(p)
        )
        p.wait()
    else:
        raise RuntimeError('invalid command: %s' % args.command)


class ExitCodes:
    UNO_CONNECTION_ERROR = 9


def main(argv: Optional[Sequence] = None):
    args = parse_args(argv)
    try:
        run(args)
    except NoConnectException as e:
        LOGGER.error('failed to connect to uno service: %s', e, exc_info=e)
        sys.exit(ExitCodes.UNO_CONNECTION_ERROR)
    except Exception as e:
        LOGGER.error('failed to to run: %s (%s)', e, type(e), exc_info=e)
        raise


if __name__ == '__main__':
    logging.basicConfig(level='INFO')

    main()



import logging

from sciencebeam_parser.models.data import (
    DocumentFeaturesContext
)
from sciencebeam_parser.models.figure.training_data import (
    FigureTeiTrainingDataGenerator,
    FigureTrainingTeiParser
)
from sciencebeam_parser.models.model import Model
from sciencebeam_parser.models.fulltext.data import FullTextDataGenerator
from sciencebeam_parser.models.figure.extract import FigureSemanticExtractor


LOGGER = logging.getLogger(__name__)


class FigureModel(Model):
    def get_data_generator(
        self,
        document_features_context: DocumentFeaturesContext
    ) -> FullTextDataGenerator:
        return FullTextDataGenerator(
            document_features_context=document_features_context
        )

    def get_semantic_extractor(self) -> FigureSemanticExtractor:
        return FigureSemanticExtractor()

    def get_tei_training_data_generator(self) -> FigureTeiTrainingDataGenerator:
        return FigureTeiTrainingDataGenerator()

    def get_training_tei_parser(self) -> FigureTrainingTeiParser:
        return FigureTrainingTeiParser()



import logging
import os
import stat
from typing import Optional

from sciencebeam_parser.utils.background_process import exec_with_logging


LOGGER = logging.getLogger(__name__)


class PdfAltoWrapper:
    def __init__(self, binary_path: str):
        self.binary_path = binary_path

    def ensure_executable(self):
        st = os.stat(self.binary_path)
        os.chmod(self.binary_path, st.st_mode | stat.S_IEXEC)

    def get_command(
        self,
        pdf_path: str,
        output_path: str,
        first_page: Optional[int] = None,
        last_page: Optional[int] = None
    ):
        command = [
            self.binary_path,
            '-noImageInline',
            '-fullFontName',
            '-noLineNumbers'
        ]
        if first_page:
            command.extend(['-f', str(first_page)])
        if last_page:
            command.extend(['-l', str(last_page)])
        command.extend([
            pdf_path,
            output_path
        ])
        return command

    def convert_pdf_to_pdfalto_xml(self,  *args, **kwargs):
        command = self.get_command(*args, **kwargs)
        LOGGER.info('command: %s', command)
        LOGGER.info('command str: %s', ' '.join(command))
        with exec_with_logging(command, logging_prefix='pdfalto') as _:
            pass



import logging

from lxml import etree

from sciencebeam_parser.utils.xml import get_text_content
from sciencebeam_parser.lookup import TextLookUp, SimpleTextLookUp


LOGGER = logging.getLogger(__name__)


TEI_NS = 'http://www.tei-c.org/ns/1.0'
TEI_NS_PREFIX = '{%s}' % TEI_NS

TEI_NS_MAP = {
    'tei': TEI_NS
}

TEI_CELL = TEI_NS_PREFIX + 'cell'


def load_xml_lookup_from_file(
    filename: str
) -> TextLookUp:
    root = etree.parse(filename)
    valid_texts = {
        get_text_content(node)
        for node in root.xpath('//tei:cell', namespaces=TEI_NS_MAP)
    }
    LOGGER.debug('valid_texts: %s', valid_texts)
    return SimpleTextLookUp(valid_texts)



import logging
from typing import (
    Dict,
    List,
    Mapping,
    Optional,
    Sequence,
    Set,
    Union
)

from lxml import etree

from sciencebeam_parser.document.semantic_document import (
    SemanticAddressField,
    SemanticAffiliationAddress,
    SemanticAuthor,
    SemanticMarker
)
from sciencebeam_parser.document.tei.common import (
    TEI_E,
    XML_ID
)
from sciencebeam_parser.document.tei.factories import (
    TeiElementFactoryContext
)


LOGGER = logging.getLogger(__name__)


def _get_tei_raw_affiliation_element_for_semantic_affiliation_address(
    semantic_affiliation_address: SemanticAffiliationAddress,
    context: TeiElementFactoryContext
) -> etree.ElementBase:
    children: List[Union[str, dict, etree.ElementBase]] = []
    children.append({'type': 'raw_affiliation'})
    pending_whitespace: str = ''
    for semantic_content in semantic_affiliation_address:
        merged_block = semantic_content.merged_block
        if pending_whitespace:
            children.append(pending_whitespace)
        if isinstance(semantic_content, SemanticMarker):
            children.append(TEI_E(
                'label',
                *context.iter_layout_block_tei_children(merged_block, enable_coordinates=False)
            ))
            pending_whitespace = merged_block.whitespace
            continue
        children.extend(
            context.iter_layout_block_tei_children(merged_block, enable_coordinates=False)
        )
        pending_whitespace = merged_block.whitespace
    return TEI_E('note', *children)


def get_tei_affiliation_for_semantic_affiliation_address_element(
    semantic_affiliation_address: SemanticAffiliationAddress,
    context: TeiElementFactoryContext
) -> etree.ElementBase:
    LOGGER.debug('semantic_affiliation_address: %s', semantic_affiliation_address)
    raw_affiliation = _get_tei_raw_affiliation_element_for_semantic_affiliation_address(
        semantic_affiliation_address,
        context=context
    )
    attributes = context.get_default_attributes_for_semantic_content(
        semantic_affiliation_address
    )
    if semantic_affiliation_address.content_id:
        attributes = {**attributes, 'key': semantic_affiliation_address.content_id}
        if XML_ID in attributes:
            del attributes[XML_ID]
    children = [
        attributes,
        raw_affiliation
    ]
    address_semantic_content_list = []
    for semantic_content in semantic_affiliation_address:
        if isinstance(semantic_content, SemanticAddressField):
            address_semantic_content_list.append(semantic_content)
            continue
        children.extend(context.get_tei_child_elements_for_semantic_content(
            semantic_content
        ))
    LOGGER.debug('address_semantic_content_list: %r', address_semantic_content_list)
    if address_semantic_content_list:
        children.append(TEI_E('address', *[
            child
            for semantic_content in address_semantic_content_list
            for child in context.get_tei_child_elements_for_semantic_content(
                semantic_content
            )
        ]))
    return TEI_E('affiliation', *children)


def get_tei_author_for_semantic_author_element(
    semantic_author: SemanticAuthor,
    context: TeiElementFactoryContext,
    affiliations_by_marker: Optional[Mapping[str, Sequence[SemanticAffiliationAddress]]] = None
) -> etree.ElementBase:
    if affiliations_by_marker is None:
        affiliations_by_marker = {}
    LOGGER.debug('semantic_author: %s', semantic_author)
    pers_name_children = []
    for semantic_content in semantic_author:
        pers_name_children.extend(context.get_tei_child_elements_for_semantic_content(
            semantic_content
        ))
    children = [
        TEI_E(
            'persName',
            context.get_default_attributes_for_semantic_content(semantic_author),
            *pers_name_children
        )
    ]
    affiliations = []
    for marker_text in semantic_author.view_by_type(SemanticMarker).get_text_list():
        semantic_affiliations = affiliations_by_marker.get(marker_text)
        if not semantic_affiliations:
            LOGGER.warning('affiliation not found for marker: %r', marker_text)
            continue
        for semantic_affiliation in semantic_affiliations:
            affiliations.append(get_tei_affiliation_for_semantic_affiliation_address_element(
                semantic_affiliation,
                context=context
            ))
    children.extend(affiliations)
    return TEI_E('author', *children)


def get_dummy_tei_author_for_semantic_affiliations_element(
    semantic_affiliations: Sequence[SemanticAffiliationAddress],
    context: TeiElementFactoryContext
) -> etree.ElementBase:
    children = [
        TEI_E('note', {'type': 'dummy_author'}, 'Dummy author for orphan affiliations')
    ]
    children.extend([
        get_tei_affiliation_for_semantic_affiliation_address_element(
            semantic_affiliation,
            context=context
        )
        for semantic_affiliation in semantic_affiliations
    ])
    return TEI_E('author', *children)


def get_authors_affiliation_markers(authors: List[SemanticAuthor]) -> Set[str]:
    return {
        marker
        for author in authors
        for marker in author.view_by_type(SemanticMarker).get_text_list()
    }


def get_orphan_affiliations(
    affiliations_by_marker: Dict[str, List[SemanticAffiliationAddress]],
    authors: List[SemanticAuthor]
) -> List[SemanticAffiliationAddress]:
    used_affiliation_markers = get_authors_affiliation_markers(authors)
    return [
        affiliation
        for marker, affiliations in affiliations_by_marker.items()
        if marker not in used_affiliation_markers
        for affiliation in affiliations
    ]



import functools

from sciencebeam_parser.cv_models.cv_model import (
    ComputerVisionModel,
    LazyComputerVisionModel
)


class EngineNames:
    LAYOUT_PARSER = 'layout_parser'


def get_layout_parser_model_for_path(config: dict, path: str) -> ComputerVisionModel:
    from sciencebeam_parser.cv_models.layout_parser_cv_model import (  # noqa pylint: disable=import-outside-toplevel
        LayoutParserComputerVisionModel
    )
    return LayoutParserComputerVisionModel(config, path)


def get_engine_name_for_config(config: dict) -> str:
    engine_name = config.get('engine')
    if engine_name:
        return engine_name
    return EngineNames.LAYOUT_PARSER


def get_cv_model_for_config(config: dict) -> ComputerVisionModel:
    path = config['path']
    engine_name = get_engine_name_for_config(config)
    if engine_name == EngineNames.LAYOUT_PARSER:
        return get_layout_parser_model_for_path(config, path)
    raise RuntimeError('invalid engine name: %r' % engine_name)


def get_lazy_cv_model_for_config(config: dict) -> ComputerVisionModel:
    return LazyComputerVisionModel(
        functools.partial(get_cv_model_for_config, config)
    )



import logging

from lxml import etree

from sciencebeam_parser.document.semantic_document import (
    SemanticContentWrapper,
    SemanticExternalIdentifier
)
from sciencebeam_parser.document.tei.common import (
    TEI_E
)
from sciencebeam_parser.document.tei.factory import (
    SingleElementTeiElementFactory,
    TeiElementFactoryContext
)


LOGGER = logging.getLogger(__name__)


class ExternalIdentifierTeiElementFactory(SingleElementTeiElementFactory):
    def get_tei_element_for_semantic_content(
        self,
        semantic_content: SemanticContentWrapper,
        context: TeiElementFactoryContext
    ) -> etree.ElementBase:
        assert isinstance(semantic_content, SemanticExternalIdentifier)
        external_identifier = semantic_content
        if LOGGER.isEnabledFor(logging.DEBUG):
            LOGGER.debug(
                'external_identifier: type=%r, value=%r, text=%r, content=%r',
                external_identifier.external_identifier_type,
                external_identifier.value,
                external_identifier.get_text(),
                external_identifier
            )
        attributes = context.get_default_attributes_for_semantic_content(external_identifier)
        if external_identifier.external_identifier_type:
            attributes = {**attributes, 'type': external_identifier.external_identifier_type}
        return TEI_E('idno', attributes, external_identifier.value)



import logging
from typing import (
    Iterable,
    List,
)

from lxml import etree

from sciencebeam_parser.document.semantic_document import (
    SemanticContentWrapper,
    SemanticFigure,
    SemanticHeading,
    SemanticLabel,
    SemanticParagraph,
    SemanticRawEquation,
    SemanticSection,
    SemanticSectionTypes,
    SemanticTable
)
from sciencebeam_parser.document.tei.common import (
    TEI_E,
    TeiElementBuilder
)
from sciencebeam_parser.document.tei.factory import (
    SingleElementTeiElementFactory,
    T_ElementChildrenList,
    TeiElementFactory,
    TeiElementFactoryContext
)


LOGGER = logging.getLogger(__name__)


class HeadingTeiElementFactory(SingleElementTeiElementFactory):
    def get_tei_element_for_semantic_content(
        self,
        semantic_content: SemanticContentWrapper,
        context: TeiElementFactoryContext
    ) -> etree.ElementBase:
        LOGGER.debug('semantic_content: %s', semantic_content)
        assert isinstance(semantic_content, SemanticHeading)
        semantic_heading = semantic_content
        children: T_ElementChildrenList = [
            context.get_default_attributes_for_semantic_content(semantic_heading)
        ]
        pending_whitespace = ''
        for child_semantic_content in semantic_heading:
            if isinstance(child_semantic_content, SemanticLabel):
                children.append({'n': child_semantic_content.get_text()})
                continue
            layout_block = child_semantic_content.merged_block
            if pending_whitespace:
                children.append(pending_whitespace)
            children.extend(context.iter_layout_block_tei_children(
                layout_block=layout_block,
                enable_coordinates=False
            ))
            pending_whitespace = layout_block.whitespace
        return TEI_E('head', *children)


def iter_flat_paragraph_formula(
    semantic_paragraph: SemanticParagraph
) -> Iterable[SemanticContentWrapper]:
    pending_semantic_content_list: List[SemanticContentWrapper] = []
    for semantic_content in semantic_paragraph:
        if isinstance(semantic_content, SemanticRawEquation):
            if pending_semantic_content_list:
                yield SemanticParagraph(pending_semantic_content_list)
                pending_semantic_content_list = []
            yield semantic_content
            continue
        pending_semantic_content_list.append(semantic_content)
    if pending_semantic_content_list:
        yield SemanticParagraph(pending_semantic_content_list)


class ParagraphTeiElementFactory(TeiElementFactory):
    def get_tei_children_for_semantic_content(
        self,
        semantic_content: SemanticContentWrapper,
        context: TeiElementFactoryContext
    ) -> List[etree.ElementBase]:
        LOGGER.debug('semantic_content: %s', semantic_content)
        assert isinstance(semantic_content, SemanticParagraph)
        semantic_paragraph = semantic_content
        result: List[etree.ElementBase] = []
        for flat_parent_semantic_content in iter_flat_paragraph_formula(semantic_paragraph):
            if not isinstance(flat_parent_semantic_content, SemanticParagraph):
                result.extend(context.get_tei_child_elements_for_semantic_content(
                    flat_parent_semantic_content
                ))
                continue
            children: T_ElementChildrenList = [
                context.get_default_attributes_for_semantic_content(flat_parent_semantic_content)
            ]
            pending_whitespace = ''
            for child_semantic_content in flat_parent_semantic_content:
                pending_whitespace = context.append_tei_children_list_and_get_whitespace(
                    children,
                    child_semantic_content,
                    pending_whitespace=pending_whitespace
                )
            result.append(TEI_E('p', *children))
        return result


class SectionTeiElementFactory(TeiElementFactory):
    def get_tei_children_for_semantic_content(
        self,
        semantic_content: SemanticContentWrapper,
        context: TeiElementFactoryContext
    ) -> List[etree.ElementBase]:
        LOGGER.debug('semantic_content: %s', semantic_content)
        assert isinstance(semantic_content, SemanticSection)
        semantic_section = semantic_content
        tei_section = TeiElementBuilder(TEI_E('div'))
        for child_semantic_content in semantic_section:
            if isinstance(child_semantic_content, (SemanticFigure, SemanticTable,)):
                # rendered at parent level
                continue
            tei_section.extend(context.get_tei_child_elements_for_semantic_content(
                child_semantic_content
            ))
        if semantic_content.section_type == SemanticSectionTypes.ACKNOWLEDGEMENT:
            tei_section.element.attrib['type'] = 'acknowledgement'
        if not list(tei_section.element):
            return []
        return [tei_section.element]



from typing import (
    Any,
    Mapping,
    Type
)

from lxml import etree

from sciencebeam_parser.document.semantic_document import (
    SemanticCitation,
    SemanticContentWrapper,
    SemanticFigureCitation,
    SemanticReferenceCitation,
    SemanticTableCitation
)
from sciencebeam_parser.document.tei.common import (
    TEI_E
)
from sciencebeam_parser.document.tei.factory import (
    SingleElementTeiElementFactory,
    TeiElementFactoryContext
)


CITATION_TYPE_BY_SEMANTIC_CLASS: Mapping[Type[Any], str] = {
    SemanticFigureCitation: 'figure',
    SemanticTableCitation: 'table',
    SemanticReferenceCitation: 'bibr'
}


class CitationTeiElementFactory(SingleElementTeiElementFactory):
    def get_tei_element_for_semantic_content(
        self,
        semantic_content: SemanticContentWrapper,
        context: TeiElementFactoryContext
    ) -> etree.ElementBase:
        assert isinstance(semantic_content, SemanticCitation)
        citation = semantic_content
        citation_type = CITATION_TYPE_BY_SEMANTIC_CLASS.get(type(citation))
        attributes = {}
        if citation_type:
            attributes['type'] = citation_type
        if citation.target_content_id:
            attributes['target'] = '#' + citation.target_content_id
        return TEI_E(
            'ref',
            attributes,
            *context.iter_layout_block_tei_children(citation.merged_block)
        )



from abc import abstractmethod
from typing import Callable, Generic, Optional, TypeVar


T = TypeVar('T')


class Preloadable:
    @abstractmethod
    def preload(self):
        raise NotImplementedError()


class LazyLoaded(Generic[T]):
    def __init__(self, factory: Callable[[], T]):
        super().__init__()
        self.factory = factory
        self._value: Optional[T] = None

    def __repr__(self) -> str:
        return '%s(_value=%r)' % (
            type(self).__name__, self._value
        )

    @property
    def is_loaded(self) -> bool:
        return self._value is not None

    def get(self) -> T:
        if self._value is None:
            self._value = self.factory()
            assert self._value is not None
        return self._value



from typing import Dict, List

from lxml import etree

from sciencebeam_parser.document.layout_document import (
    LayoutGraphic,
    LayoutLineMeta,
    LayoutPageCoordinates,
    LayoutFont,
    LayoutPageMeta,
    LayoutToken,
    LayoutLine,
    LayoutBlock,
    LayoutPage,
    LayoutDocument,
    EMPTY_FONT
)


ALTO_NS = 'http://www.loc.gov/standards/alto/ns-v3#'
ALTO_NS_MAP = {
    'alto': ALTO_NS
}


def alto_xpath(parent: etree.ElementBase, xpath: str) -> List[etree.ElementBase]:
    return parent.xpath(xpath, namespaces=ALTO_NS_MAP)


class AltoParser:
    def __init__(self):
        self.font_by_id_map: Dict[str, LayoutFont] = {}

    def parse_page_coordinates(
        self,
        node: etree.ElementBase,
        page_number: int
    ) -> LayoutPageCoordinates:
        return LayoutPageCoordinates(
            x=float(node.attrib.get('HPOS', 0)),
            y=float(node.attrib.get('VPOS', 0)),
            width=float(node.attrib.get('WIDTH', 0)),
            height=float(node.attrib.get('HEIGHT', 0)),
            page_number=page_number
        )

    def parse_token(
        self,
        token_node: etree.ElementBase,
        page_number: int,
        layout_line_meta: LayoutLineMeta
    ) -> LayoutToken:
        return LayoutToken(
            text=token_node.attrib.get('CONTENT') or '',
            font=self.font_by_id_map.get(
                token_node.attrib.get('STYLEREFS'),
                EMPTY_FONT
            ),
            coordinates=self.parse_page_coordinates(token_node, page_number=page_number),
            line_meta=layout_line_meta
        )

    def parse_line(
        self,
        line_node: etree.ElementBase,
        page_number: int,
        page_meta: LayoutPageMeta
    ) -> LayoutLine:
        return LayoutLine(tokens=[
            self.parse_token(
                token_node,
                page_number=page_number,
                layout_line_meta=LayoutLineMeta(
                    line_id=id(line_node),
                    page_meta=page_meta
                )
            )
            for token_node in alto_xpath(line_node, './/alto:String')
        ])

    def parse_block(
        self,
        block_node: etree.ElementBase,
        page_number: int,
        page_meta: LayoutPageMeta
    ) -> LayoutBlock:
        return LayoutBlock(lines=[
            self.parse_line(line_node, page_number=page_number, page_meta=page_meta)
            for line_node in alto_xpath(block_node, './/alto:TextLine[alto:String]')
        ])

    def parse_graphic(
        self,
        graphic_node: etree.ElementBase,
        page_number: int,
        page_meta: LayoutPageMeta
    ) -> LayoutGraphic:
        attrib = graphic_node.attrib
        return LayoutGraphic(
            local_file_path=attrib.get('FILEID'),
            coordinates=self.parse_page_coordinates(graphic_node, page_number=page_number),
            graphic_type=attrib.get('TYPE'),
            page_meta=page_meta
        )

    def parse_page(
        self,
        page_node: etree.ElementBase,
        page_index: int
    ) -> LayoutPage:
        page_number_str = page_node.attrib.get('PHYSICAL_IMG_NR')
        page_number = int(page_number_str) if page_number_str else 1 + page_index
        width_str = page_node.attrib.get('WIDTH')
        height_str = page_node.attrib.get('HEIGHT')
        coordinates = (
            LayoutPageCoordinates(
                x=0,
                y=0,
                width=float(width_str),
                height=float(height_str),
                page_number=page_number
            )
            if width_str and height_str
            else None
        )
        page_meta = LayoutPageMeta(
            page_number=page_number,
            coordinates=coordinates
        )
        return LayoutPage(
            meta=page_meta,
            blocks=[
                self.parse_block(block_node, page_number=page_number, page_meta=page_meta)
                for block_node in alto_xpath(page_node, './/alto:TextBlock')
            ],
            graphics=[
                self.parse_graphic(graphic_node, page_number=page_number, page_meta=page_meta)
                for graphic_node in alto_xpath(page_node, './/alto:Illustration')
            ]
        )

    def parse_font(self, font_node: etree.ElementBase) -> LayoutFont:
        font_styles = (font_node.attrib.get('FONTSTYLE') or '').split(' ')
        return LayoutFont(
            font_id=font_node.attrib.get('ID'),
            font_family=font_node.attrib.get('FONTFAMILY'),
            font_size=float(font_node.attrib.get('FONTSIZE')),
            is_bold='bold' in font_styles,
            is_italics='italics' in font_styles,
            is_subscript='subscript' in font_styles,
            is_superscript='superscript' in font_styles
        )

    def parse_font_by_id_map(self, root: etree.ElementBase) -> Dict[str, LayoutFont]:
        fonts = [
            self.parse_font(font_node)
            for font_node in alto_xpath(root, './alto:Styles/alto:TextStyle')
        ]
        return {
            font.font_id: font
            for font in fonts
        }

    def parse_root(self, root: etree.ElementBase) -> LayoutDocument:
        self.font_by_id_map = self.parse_font_by_id_map(root)
        return LayoutDocument(pages=[
            self.parse_page(page_node, page_index=page_index)
            for page_index, page_node in enumerate(alto_xpath(root, './/alto:Page'))
        ])


def parse_alto_root(root: etree.ElementBase) -> LayoutDocument:
    return AltoParser().parse_root(root)



from typing import Iterable

from sciencebeam_parser.models.data import (
    ContextAwareLayoutTokenFeatures,
    ContextAwareLayoutTokenModelDataGenerator,
    LayoutModelData
)


class FullTextDataGenerator(ContextAwareLayoutTokenModelDataGenerator):
    def iter_model_data_for_context_layout_token_features(
        self,
        token_features: ContextAwareLayoutTokenFeatures
    ) -> Iterable[LayoutModelData]:
        yield token_features.get_layout_model_data([
            token_features.token_text,
            token_features.get_lower_token_text(),
            token_features.get_prefix(1),
            token_features.get_prefix(2),
            token_features.get_prefix(3),
            token_features.get_prefix(4),
            token_features.get_suffix(1),
            token_features.get_suffix(2),
            token_features.get_suffix(3),
            token_features.get_suffix(4),
            token_features.get_block_status_with_blockstart_for_single_token(),
            token_features.get_line_status_with_linestart_for_single_token(),
            token_features.get_alignment_status(),
            token_features.get_token_font_status(),
            token_features.get_token_font_size_feature(),
            token_features.get_str_is_bold(),
            token_features.get_str_is_italic(),
            token_features.get_capitalisation_status_using_allcap(),
            token_features.get_digit_status_using_containsdigits(),
            token_features.get_str_is_single_char(),
            token_features.get_punctuation_type_feature(),
            token_features.get_dummy_str_relative_document_position(),
            token_features.get_dummy_str_relative_page_position(),
            token_features.get_dummy_str_is_bitmap_around(),
            token_features.get_dummy_callout_type(),
            token_features.get_dummy_str_is_callout_known(),
            token_features.get_str_is_superscript()
        ])



from typing import List, Sequence, Union

from lxml import etree


def get_text_content(node: Union[str, etree.ElementBase]) -> str:
    if node is None:
        return ''
    if isinstance(node, str):
        return str(node)
    return ''.join(node.itertext())


def get_text_content_list(nodes: Sequence[Union[str, etree.ElementBase]]) -> List[str]:
    return [
        get_text_content(node)
        for node in nodes
    ]
